#
# parakeet.sample.py
#
# Copyright (C) 2019 Diamond Light Source and Rosalind Franklin Institute
#
# Author: James Parkhurst
#
# This code is distributed under the GPLv3 license, a copy of
# which is included in the root directory of this package.
#
from collections import defaultdict
import h5py
import gemmi
import logging
import numpy as np
import pandas
import scipy.constants
import time
from math import pi, sqrt, floor, ceil
from scipy.spatial.transform import Rotation
import parakeet.data
import parakeet.freeze

try:
    import multem
except ImportError:
    pass

# np.random.seed(0)

# Get the logger
logger = logging.getLogger(__name__)


def get_atom_sigma_sq(atom):
    """
    Get the sigma_sq from the atom

    Args:
        atom (object): A Gemmi atom object

    Returns:
        float: The positional sigma (sqrt(b_iso))

    """
    b_iso = atom.b_iso
    return b_iso / (8 * pi**2)


def get_atom_sigma(atom):
    """
    Get the sigma from the atom

    Args:
        atom (object): A Gemmi atom object

    Returns:
        float: The positional sigma (sqrt(b_iso))

    """
    return sqrt(get_atom_sigma_sq(atom))


def translate(atom_data, translation):
    """
    Helper function to translate atom data

    Args:
        atom_data (object): The atom data
        translation (array): The translation

    """
    coords = atom_data[["x", "y", "z"]].to_numpy()
    coords += np.array(translation, dtype=coords.dtype)
    return atom_data.assign(x=coords[:, 0], y=coords[:, 1], z=coords[:, 2])


def recentre(atom_data, position=None):
    """
    Helper function to reposition atom data

    Args:
        atom_data (object): The atom data
        translation (array): The translation

    """
    # Check the input
    if position is None:
        position = (0, 0, 0)

    # Get the coords
    coords = atom_data[["x", "y", "z"]].to_numpy()

    # Compute the translation
    translation = np.array(position, dtype=coords.dtype) - coords.mean(axis=0)

    # Do the translation
    return translate(atom_data, translation)


def number_of_water_molecules(volume, density=940.0):
    """
    Compute the number of water molecules

    Args:
        volume (float): The volume A^3
        density (float): The density Kg/m^3

    Returns:
        The number of water molcules

    """

    # Determine the number of waters to place
    avogadros_number = scipy.constants.Avogadro
    molar_mass_of_water = 18.01528  # grams / mole
    density_of_water = density  # kg / m^3
    mass_of_water = (density_of_water * 1000) * (volume * 1e-10**3)  # g
    return int(floor((mass_of_water / molar_mass_of_water) * avogadros_number))


def random_uniform_rotation(size=1):
    """
    Return a uniform rotation vector sampled on a sphere

    Args:
        size (int): The number of vectors

    Returns:
        vector: The rotation vector

    """
    u1 = np.random.uniform(0, 1, size=size)
    u2 = np.random.uniform(0, 1, size=size)
    u3 = np.random.uniform(0, 1, size=size)
    theta = np.arccos(2 * u1 - 1)
    phi = 2 * pi * u2
    x = np.sin(theta) * np.cos(phi)
    y = np.sin(theta) * np.sin(phi)
    z = np.cos(theta)
    vector = np.array((x, y, z)).T
    vector *= 2 * pi * u3.reshape((u3.size, 1))
    return vector


def distribute_boxes_uniformly(volume_box, boxes, max_tries=1000):
    """
    Find n random non overlapping positions for cuboids within a volume

    Args:
        volume_box (array): The size of the volume
        boxes (array): The list of boxes
        max_tries (int): The maximum tries per cube

    Returns:
        list: A list of centre positions

    """

    # Cast to numpy array
    volume_lower, volume_upper = np.array(volume_box)
    boxes = np.array(boxes)

    # Check if the cube overlaps with any other
    def overlapping(positions, box_sizes, q, q_size):
        for p, p_size in zip(positions, box_sizes):
            p0 = p - p_size / 2
            p1 = p + p_size / 2
            q0 = q - q_size / 2
            q1 = q + q_size / 2
            if not (
                q0[0] > p1[0]
                or q1[0] < p0[0]
                or q0[1] > p1[1]
                or q1[1] < p0[1]
                or q0[2] > p1[2]
                or q1[2] < p0[2]
            ):
                return True
        return False

    # Loop until we have added the boxes
    positions = []
    box_sizes = []
    for box_size in boxes:

        # The bounds to search in
        lower = volume_lower + box_size / 2
        upper = volume_upper - box_size / 2
        assert lower[0] < upper[0]
        assert lower[1] < upper[1]
        assert lower[2] < upper[2]

        # Try to add the box
        num_tries = 0
        while True:
            q = np.random.uniform(lower, upper)
            if len(positions) == 0 or not overlapping(
                positions, box_sizes, q, box_size
            ):
                positions.append(q)
                box_sizes.append(box_size)
                break
            num_tries += 1
            if num_tries > max_tries:
                print("Number of particles placed = %d" % (len(box_sizes)))
                raise RuntimeError(f"Unable to place cube of size {box_size}")

    # Return the cube positions
    return positions


def shape_bounding_box(centre, shape):
    """
    Compute the shape bounding box

    Args:
        centre (array): The centre of the shape
        shape (object): The shape

    Returns:
        array: bounding box

    """

    def cube_bounding_box(cube):
        length = cube["length"]
        return ((0, 0, 0), (length, length, length))

    def cuboid_bounding_box(cuboid):
        length_x = cuboid["length_x"]
        length_y = cuboid["length_y"]
        length_z = cuboid["length_z"]
        return ((0, 0, 0), (length_x, length_y, length_z))

    def cylinder_bounding_box(cylinder):
        length = cylinder["length"]
        radius = np.mean(cylinder["radius"])
        return ((0, 0, 0), (2 * radius, length, 2 * radius))

    # The bounding box
    x0, x1 = np.array(
        {
            "cube": cube_bounding_box,
            "cuboid": cuboid_bounding_box,
            "cylinder": cylinder_bounding_box,
        }[shape["type"]](shape[shape["type"]])
    )

    # The offset
    offset = centre - (x1 + x0) / 2.0

    # Return the bounding box
    return (x0 + offset, x1 + offset)


def shape_bounding_cylinder(centre, shape):
    """
    Compute the shape cylinder

    Args:
        centre (array): The centre of the shape
        shape (object): The shape

    Returns:
        array: bounding cylinder

    """

    def cube_bounding_cylinder(cube):
        length = cube["length"]
        return (length, sqrt(length**2 + length**2))

    def cuboid_bounding_cylinder(cuboid):
        length_x = cuboid["length_x"]
        length_y = cuboid["length_y"]
        length_z = cuboid["length_z"]
        return (length_x, sqrt(length_y**2 + length_z**2))

    def cylinder_bounding_cylinder(cylinder):
        length = cylinder["length"]
        radius = cylinder["radius"]
        return (length, radius)

    # The bounding box
    length, radius = {
        "cube": cube_bounding_cylinder,
        "cuboid": cuboid_bounding_cylinder,
        "cylinder": cylinder_bounding_cylinder,
    }[shape["type"]](shape[shape["type"]])

    # Return the bounding box
    return (centre, length, radius)


def shape_enclosed_box(centre, shape):
    """
    Compute the shape enclosed box

    Args:
        centre (array): The centre of the shape
        shape (object): The shape

    Returns:
        array: enclosed box

    """

    def cube_enclosed_box(cube):
        length = cube["length"]
        return ((0, 0, 0), (length, length, length))

    def cuboid_enclosed_box(cuboid):
        length_x = cuboid["length_x"]
        length_y = cuboid["length_y"]
        length_z = cuboid["length_z"]
        return ((0, 0, 0), (length_x, length_y, length_z))

    def cylinder_enclosed_box(cylinder):
        length = cylinder["length"]
        radius = np.mean(cylinder["radius"])
        return (
            (radius * (1 - 1 / sqrt(2)), 0, radius * (1 - 1 / sqrt(2))),
            (radius * (1 + 1 / sqrt(2)), length, radius * (1 + 1 / sqrt(2))),
        )

    # The enclosed box
    x0, x1 = np.array(
        {
            "cube": cube_enclosed_box,
            "cuboid": cuboid_enclosed_box,
            "cylinder": cylinder_enclosed_box,
        }[shape["type"]](shape[shape["type"]])
    )

    # The offset
    offset = centre - (x1 + x0) / 2.0

    # The margin
    margin = np.array(shape.get("margin", (0, 0, 0)))

    # Return the bounding box
    return (x0 + offset + margin, x1 + offset - margin)


def is_shape_inside_box(box, centre, shape):
    """
    Check if the shape is inside the bounding box

    Args:
        box (array): The bounding box
        centre (array): The centre of the shape
        shape (object): The shape

    Returns:
        bool: Is the shape inside the box

    """

    # Get the shape bounding box
    (x0, y0, z0), (x1, y1, z1) = shape_bounding_box(centre, shape)

    # Check the bounding box is inside the containing box
    return (
        x0 >= 0
        and y0 >= 0
        and z0 >= 0
        and x1 <= box[0]
        and y1 <= box[1]
        and z1 <= box[2]
    )


def is_box_inside_shape(box, centre, shape):
    """
    Check if the box is inside the shape

    Args:
        box (array): The bounding box
        centre (array): The centre of the shape
        shape (object): The shape

    Returns:
        bool: Is the box is inside shape

    """

    def is_box_inside_cube(x0, x1, centre, cube):
        length = cube["length"]
        return (
            x0[0] >= centre[0] - length / 2.0
            and x0[1] >= centre[1] - length / 2.0
            and x0[2] >= centre[2] - length / 2.0
            and x1[0] <= centre[0] + length / 2.0
            and x1[1] <= centre[1] + length / 2.0
            and x1[2] <= centre[2] + length / 2.0
        )

    def is_box_inside_cuboid(x0, x1, centre, cuboid):
        length_x = cuboid["length_x"]
        length_y = cuboid["length_y"]
        length_z = cuboid["length_z"]
        return (
            x0[0] >= centre[0] - length_x / 2.0
            and x0[1] >= centre[1] - length_y / 2.0
            and x0[2] >= centre[2] - length_z / 2.0
            and x1[0] <= centre[0] + length_x / 2.0
            and x1[1] <= centre[1] + length_y / 2.0
            and x1[2] <= centre[2] + length_z / 2.0
        )

    def is_box_inside_cylinder(x0, x1, centre, cylinder):
        length = cylinder["length"]
        radius = np.mean(cylinder["radius"])
        return (
            (x0[0] - centre[0]) ** 2 + (x0[2] - centre[2]) ** 2 <= radius**2
            and (x1[0] - centre[0]) ** 2 + (x0[2] - centre[2]) ** 2 <= radius**2
            and (x0[0] - centre[0]) ** 2 + (x1[2] - centre[2]) ** 2 <= radius**2
            and (x1[0] - centre[0]) ** 2 + (x1[2] - centre[2]) ** 2 <= radius**2
            and x0[1] >= centre[1] - length / 2.0
            and x1[1] < centre[1] + length / 2.0
        )

    return {
        "cube": is_box_inside_cube,
        "cuboid": is_box_inside_cuboid,
        "cylinder": is_box_inside_cylinder,
    }[shape["type"]](box[0], box[1], centre, shape[shape["type"]])


class AtomData(object):
    """
    A class to hold the atom data

    """

    # The column data
    column_data = {
        "atomic_number": "uint8",
        "x": "float32",
        "y": "float32",
        "z": "float32",
        "sigma": "float32",
        "occupancy": "float32",
        "charge": "uint8",
    }

    def __init__(self, data=None, **kwargs):
        """
        Initialise the class

        Either the data attribute or the individual arrays must be set.

        Args:
            data (object): The atom data table
            atomic_number (array): The array of atomic numbers
            x (array): The array of x positions
            y (array): The array of y positions
            z (array): The array of z positions
            sigma (array): The array of sigmas
            occupancy (array): The array of occupancies
            charge (array): The array of charges

        """
        if data is None:
            self.data = pandas.DataFrame(
                dict(
                    (name, pandas.Series(kwargs[name], dtype=dtype))
                    for name, dtype in AtomData.column_data.items()
                )
            )
        else:
            if len(data) > 0:
                column_names = list(data.keys())
                for name in AtomData.column_data:
                    assert name in column_names
            self.data = data

    def rotate(self, vector):
        """
        Rotate the atom data around the rotation vector

        Args:
            vector (array): The rotation vector

        """
        if len(self.data) > 0:
            coords = self.data[["x", "y", "z"]].to_numpy()
            coords = Rotation.from_rotvec(vector).apply(coords).astype(coords.dtype)
            self.data["x"] = coords[:, 0]
            self.data["y"] = coords[:, 1]
            self.data["z"] = coords[:, 2]
        return self

    def translate(self, translation):
        """
        Translate the atom data

        Args:
            translation (array): The translation

        """
        if len(self.data) > 0:
            coords = self.data[["x", "y", "z"]].to_numpy()
            coords += np.array(translation, dtype=coords.dtype)
            self.data["x"] = coords[:, 0]
            self.data["y"] = coords[:, 1]
            self.data["z"] = coords[:, 2]
        return self

    def to_multem(self):
        """
        Convert to a multem atom list

        Returns:
            object: A multem atom list

        """
        if len(self.data) == 0:
            return multem.AtomList()
        return multem.AtomList(
            zip(
                self.data["atomic_number"].astype("uint8"),
                self.data["x"].astype("float32"),
                self.data["y"].astype("float32"),
                self.data["z"].astype("float32"),
                self.data["sigma"].astype("float32"),
                [float(1) for i in range(self.data.shape[0])],
                [int(0) for i in range(self.data.shape[0])],
                self.data["charge"].astype("uint8"),
            )
        )

    def rows(self):
        """
        Iterate through atoms rows

        Yield:
            tuple: atom data

        """
        return zip(
            self.data["atomic_number"],
            self.data["x"],
            self.data["y"],
            self.data["z"],
            self.data["sigma"],
            self.data["occupancy"],
            self.data["charge"],
        )

    @classmethod
    def from_gemmi_structure(Class, structure):
        """
        Read the sample from a gemmi sucture

        Args:
            structure (object): The input structure

        Returns:
            object: The Sample object

        """

        # The column order
        column_info = ["atomic_number", "x", "y", "z", "sigma", "occupancy", "charge"]

        # Iterate through the atoms
        def iterate_atoms(structure):
            for model_index, model in enumerate(structure):
                for chain in model:
                    for residue in chain:
                        for atom in residue:
                            assert atom.element.atomic_number > 0
                            yield (
                                atom.element.atomic_number,
                                atom.pos.x,
                                atom.pos.y,
                                atom.pos.z,
                                get_atom_sigma(atom),
                                atom.occ,
                                atom.charge,
                            )

        # Create a dictionary of column data
        def create_atom_data(structure):
            return dict(
                (name, pandas.Series(data, dtype=AtomData.column_data[name]))
                for data, name in zip(zip(*iterate_atoms(structure)), column_info)
            )

        # Return the sample with the atom data as a pandas dataframe
        return AtomData(data=pandas.DataFrame(create_atom_data(structure)))

    @classmethod
    def from_gemmi_file(Class, filename):
        """
        Read the sample from a file

        Args:
            filename (str): The input filename

        Returns:
            object: The Sample object

        """

        # Read the structure
        return Class.from_gemmi_structure(gemmi.read_structure(filename))

    @classmethod
    def from_ligand_file(Class, filename):
        """
        Read the sample from a file

        Args:
            filename (str): The input filename

        Returns:
            object: The Sample object

        """
        # Create a single ribosome sample
        return Class.from_gemmi_structure(
            gemmi.make_structure_from_chemcomp_block(gemmi.cif.read(filename)[0])
        )

    @classmethod
    def from_text_file(Class, filename):
        """
        Read the sample from a text file

        Args:
            filename (str): The input filename

        Returns:
            object: The Sample object

        """

        # The column order
        column_info = ["atomic_number", "x", "y", "z", "sigma", "occupancy", "charge"]

        # Iterate through the atoms
        def iterate_atoms(infile):
            for line in infile.readlines():
                tokens = line.split()
                yield (
                    int(tokens[0]),
                    float(tokens[1]),
                    float(tokens[2]),
                    float(tokens[3]),
                    float(tokens[4]),
                    float(tokens[5]),
                    int(tokens[6]),
                )

        # Create a dictionary of column data
        def create_atom_data(infile):
            return dict(
                (name, pandas.Series(data, dtype=AtomData.column_data[name]))
                for data, name in zip(zip(*iterate_atoms(infile)), column_info)
            )

        # Return the atom data
        with open(filename, "r") as infile:
            return AtomData(data=pandas.DataFrame(create_atom_data(infile)))


class SampleHDF5Adapter(object):
    """
    A class to handle the HDF5 group and dataset creation

    """

    class AtomsSubGroup(object):
        """
        A class to handle the atom sub groups

        """

        def __init__(self, handle, name):
            """
            Save the group handle

            Args:
                handle (object): The group handle
                name (str): The dataset name

            """

            # Save the handle
            self.__handle = handle
            self.name = name

            # Create the datasets
            if name not in self.__handle:
                dtype = [(key, value) for key, value in AtomData.column_data.items()]
                self.__handle.create_dataset(
                    name, (0,), maxshape=(None,), dtype=dtype, chunks=True
                )

        @property
        def atoms(self):
            """
            Return the atom data in the atom sub group

            """
            return pandas.DataFrame.from_records(self.__handle[self.name][:])

        @atoms.setter
        def atoms(self, data):
            """
            Set the atom data in the sub group

            """

            # Check size, resize and set data
            self.__handle[self.name].resize((len(data),))
            self.__handle[self.name][:] = data.to_records()

        def extend(self, data):
            """
            Set the atom data in the sub group

            """
            # Get the size
            i0 = self.__handle[self.name].shape[0]
            i1 = i0 + data.shape[0]

            # Check size, resize and set data
            assert i0 == self.__handle[self.name].shape[0]
            self.__handle[self.name].resize((i1,))
            self.__handle[self.name][i0:i1] = data.to_records()

        def __len__(self):
            """
            Returns the number of atoms

            """
            if len(self.__handle) == 0:
                return 0
            return self.__handle[self.name].shape[0]

    class AtomsGroup(object):
        """
        A class to hold the atom group

        """

        def __init__(self, handle):
            """
            Save the group handle

            Args:
                handle (object): The group handle

            """
            self.__handle = handle

        def __len__(self):
            """
            Returns:
                int: The number of sub groups

            """
            return len(self.__handle)

        def __getitem__(self, item):
            """
            Get the subgroup with the given name

            Args:
                item (str): The subgroup name

            Returns:
                object: The subgroup

            """
            return SampleHDF5Adapter.AtomsSubGroup(self.__handle, item)

        def __iter__(self):
            """
            Iterate through the names of the subgroups

            """
            return self.keys()

        def keys(self):
            """
            Iterate through the names of the subgroups

            """
            return self.__handle.keys()

        def items(self):
            """
            Iterate through the names and the subgroups

            """
            for key in self.__handle.keys():
                yield key, SampleHDF5Adapter.AtomsSubGroup(self.__handle, key)

        def values(self):
            """
            Iterate through the subgroups

            """
            for key in self.__handle.keys():
                yield SampleHDF5Adapter.AtomsSubGroup(self.__handle, key)

        def bins(self):
            """
            Iterate through the z bins

            """
            for key in self.keys():
                x, y, z = [int(k.split("=")[1]) for k in key.split(";")]
                yield key, np.array((x, y, z))

        def number_of_atoms(self):
            """
            Get the total number of atoms in all the subgroups

            Returns:
                int: The total number of atoms

            """
            count = 0
            for key, value in self.items():
                count += len(value)
            return count

    class MoleculeGroup(object):
        """
        A class to represent the molecule group

        """

        def __init__(self, handle):
            """
            Save the group handle

            Args:
                handle (object): The group handle

            """
            self.__handle = handle

        @property
        def atoms(self):
            """
            Get the atoms

            """
            return pandas.DataFrame.from_records(self.__handle["atoms"][:])

        @property
        def positions(self):
            """
            Get the positions

            """
            return self.__handle["positions"][:]

        @property
        def orientations(self):
            """
            Get the orientations

            """
            return self.__handle["orientations"][:]

        @atoms.setter
        def atoms(self, data):
            """
            Set the atom data

            """
            dtype = [(key, value) for key, value in AtomData.column_data.items()]
            self.__handle.create_dataset(
                "atoms", data=data.to_records(), dtype=dtype, chunks=True
            )

        @positions.setter
        def positions(self, positions):
            """
            Set the positions

            """
            self.__handle.create_dataset("positions", data=positions, dtype="float32")

        @orientations.setter
        def orientations(self, orientations):
            """
            Set the orientations

            """
            self.__handle.create_dataset(
                "orientations", data=orientations, dtype="float32"
            )

    class MoleculeListGroup(object):
        """
        A class to represent the molecule list group

        """

        def __init__(self, handle):
            """
            Save the group handle

            Args:
                handle (object): The group handle

            """
            self.__handle = handle

        def __len__(self):
            """
            Return the number of molecules

            """
            return len(self.__handle)

        def __getitem__(self, item):
            """
            Get the molecule by name

            """
            if item not in self.__handle:
                self.__handle.create_group(item)
            return SampleHDF5Adapter.MoleculeGroup(self.__handle[item])

        def __iter__(self):
            """
            Iterate through the molecule names

            """
            return self.keys()

        def keys(self):
            """
            Iterate through the molecule names

            """
            return self.__handle.keys()

        def items(self):
            """
            Iterate through the molecule names and groups

            """
            for key, value in self.__handle.items():
                yield key, SampleHDF5Adapter.MoleculeGroup(value)

        def values(self):
            """
            Iterate through the molecule groups

            """
            for value in self.__handle.values():
                yield SampleHDF5Adapter.MoleculeGroup(value)

    class SampleGroup(object):
        """
        A class to represent the sample group

        """

        def __init__(self, handle):
            """
            Save the group handle

            Args:
                handle (object): The group handle

            """
            self.__handle = handle

        @property
        def atoms(self):
            """
            Get the atom group

            """
            if "atoms" not in self.__handle:
                self.__handle.create_group("atoms")
            return SampleHDF5Adapter.AtomsGroup(self.__handle["atoms"])

        @property
        def molecules(self):
            """
            Get the molecule group

            """
            if "molecules" not in self.__handle:
                self.__handle.create_group("molecules")
            return SampleHDF5Adapter.MoleculeListGroup(self.__handle["molecules"])

        @property
        def bounding_box(self):
            """
            Get the bounding box

            """
            if "bounding_box" not in self.__handle:
                return np.zeros(shape=(2, 3), dtype="float32")
            return self.__handle["bounding_box"][:]

        @bounding_box.setter
        def bounding_box(self, bbox):
            """
            Set the bounding box

            """
            if "bounding_box" not in self.__handle:
                self.__handle.create_dataset("bounding_box", data=bbox, dtype="float32")
            else:
                self.__handle["bounding_box"][:] = bbox

        @property
        def containing_box(self):
            """
            Get the containing box

            """
            if "containing_box" not in self.__handle:
                return np.zeros(shape=(2, 3), dtype="float32")
            return self.__handle["containing_box"][:]

        @containing_box.setter
        def containing_box(self, box):
            """
            Set the containing box

            """
            if "containing_box" not in self.__handle:
                self.__handle.create_dataset(
                    "containing_box", data=box, dtype="float32"
                )
            else:
                self.__handle["containing_box"][:] = box

        @property
        def centre(self):
            """
            Get the centre

            """
            if "centre" not in self.__handle:
                return np.zeros(shape=(3,), dtype="float32")
            return self.__handle["centre"][:]

        @centre.setter
        def centre(self, centre):
            """
            Set the centre

            """
            if "centre" not in self.__handle:
                self.__handle.create_dataset("centre", data=centre, dtype="float32")
            else:
                self.__handle["centre"][:] = centre

        @property
        def shape(self):
            """
            Get the shape

            """
            if "shape" not in self.__handle:
                return {}

            # Get the shape
            shape = self.__handle["shape"][0]
            if isinstance(shape, bytes):
                shape = shape.decode("utf-8")
            return {
                "type": shape,
                "margin": self.__handle["shape"].attrs.get("margin", [0, 0, 0]),
                shape: dict(list(self.__handle["shape"].attrs.items())),
            }

        @shape.setter
        def shape(self, shape):
            """
            Set the shape

            """
            if "shape" not in self.__handle:
                self.__handle.create_dataset(
                    "shape", shape=(1,), dtype=h5py.string_dtype(encoding="utf-8")
                )

            # Set the shape
            self.__handle["shape"][:] = shape["type"]

            # Set the shape attributes
            for key, value in shape[shape["type"]].items():
                self.__handle["shape"].attrs[key] = value

            # Set the margin
            self.__handle["shape"].attrs["margin"] = shape.get("margin", (0, 0, 0))

    def __init__(self, filename=None, mode="r"):
        """

        Initialise the class with a filename

        Args:
            filename (str): The filename
            mode (str): The open mode

        """
        self.__handle = h5py.File(filename, mode=mode)

    @property
    def sample(self):
        """
        Get the sample group

        """
        if "sample" not in self.__handle:
            self.__handle.create_group("sample")
        return SampleHDF5Adapter.SampleGroup(self.__handle["sample"])

    def close(self):
        """
        Close the file

        """
        self.__handle.close()


class Sample(object):
    """
    A class to hold the sample data

    """

    def __init__(self, filename=None, mode="r"):
        """

        Initialise the class with a filename

        Args:
            filename (str): The filename
            mode (str): The open mode

        """

        # Open the HDF5 file
        self.__handle = SampleHDF5Adapter(filename, mode=mode)

        # The step between datasets, A 500^3 A^3 volume has around 4M water molecules
        # This seems to be a reasonable division size
        self.step = 100_000  # A

    def close(self):
        """
        Close the HDF5 file

        """
        self.__handle.close()

    def atoms_dataset_name(self, x):
        """
        Get the atom dataset name

        """
        return "X=%06d; Y=%06d; Z=%06d" % tuple(x)

    def atoms_dataset_range(self, x0, x1):
        """
        Get the atom dataset range

        Args:
            x0 (array): The minimum coord
            x1 (array): The maximum coord

        Yields:
            tuple: The coordinates of the sub ranges

        """
        x0 = (np.floor(np.array(x0) / float(self.step)) * self.step).astype("int32")
        x1 = (np.ceil(np.array(x1) / float(self.step)) * self.step).astype("int32")
        for z in range(x0[2], x1[2], self.step):
            for y in range(x0[1], x1[1], self.step):
                for x in range(x0[0], x1[0], self.step):
                    c = np.array((x, y, z))
                    yield c, c + self.step

    def atoms_dataset_range_3d(self, x0, x1):
        """
        Get the atom dataset range

        Args:
            x0 (array): The minimum coord
            x1 (array): The maximum coord

        Yields:
            tuple: The coordinates of the sub ranges

        """
        x0 = (np.floor(np.array(x0) / float(self.step)) * self.step).astype("int32")
        x1 = (np.ceil(np.array(x1) / float(self.step)) * self.step).astype("int32")
        shape = np.floor((x1 - x0) / self.step).astype("int32")
        X, Y, Z = np.mgrid[0 : shape[0], 0 : shape[1], 0 : shape[2]]
        indices = np.vstack((X.flatten(), Y.flatten(), Z.flatten())).T
        return indices, x0 + indices * self.step

    @property
    def dimensions(self):
        """
        The dimensions of the bounding box

        Returns:
            array: The dimensions

        """
        x0, x1 = self.bounding_box
        return x1 - x0

    def add_atoms(self, atoms):
        """
        Add some atoms

        Args:
            atoms (object): The atom data to add to the sample

        """

        # Get the min/max coords
        x0 = atoms.data["x"].min()
        y0 = atoms.data["y"].min()
        z0 = atoms.data["z"].min()
        x1 = atoms.data["x"].max()
        y1 = atoms.data["y"].max()
        z1 = atoms.data["z"].max()

        # Check the old bounding box
        if self.number_of_atoms > 0:

            # Get the bounding box
            (bx0, by0, bz0), (bx1, by1, bz1) = self.bounding_box

            # Get the min/max coords
            x0 = min(x0, bx0)
            y0 = min(y0, by0)
            z0 = min(z0, bz0)
            x1 = max(x1, bx1)
            y1 = max(y1, by1)
            z1 = max(z1, bz1)

        # Set the bounding box
        self.__handle.sample.bounding_box = [(x0, y0, z0), (x1, y1, z1)]

        # Get the data bounds
        coords = atoms.data[["x", "y", "z"]].to_numpy()

        # Get the grid index and coordinates
        grid_index, x_min = self.atoms_dataset_range_3d(
            coords.min(axis=0), coords.max(axis=0)
        )

        # Compute the index of each coordinate
        index = np.floor((coords - x_min[0]) / self.step).astype("int32")
        size = np.max(grid_index, axis=0) + 1
        index = index[:, 2] + index[:, 1] * size[2] + index[:, 0] * size[2] * size[1]

        # Create an index list for each subgroup by splitting the sorted indices
        sorted_index = np.argsort(index)
        split_index, splits = np.unique(index[sorted_index], return_index=True)
        index_list = np.split(sorted_index, splits[1:])

        # Ensure the number is correct
        assert sum(map(len, index_list)) == atoms.data.shape[0]
        # assert np.max(split_index) < len(x_min)

        # Add the atoms to the subgroups
        for i in range(len(split_index)):
            try:
                name = self.atoms_dataset_name(x_min[split_index[i]])
                data = atoms.data.iloc[index_list[i]]
                self.__handle.sample.atoms[name].extend(data)
            except Exception:
                print("SKIPPING: ", i)

    def del_atoms(self, deleter):
        """
        Delete atoms in the given range

        Args:
            deleter (func): Check atoms and return only those we want to keep

        """

        # Iterate over dataset ranges
        for x_min, x_max in self.atoms_dataset_range(deleter.x0, deleter.x1):
            name = self.atoms_dataset_name(x_min)
            atoms = self.__handle.sample.atoms[name].atoms
            self.__handle.sample.atoms[name].atoms = deleter(atoms)

    def add_molecule(self, atoms, positions, orientations, name=None):
        """
        Add a molecule.

        This adds the atoms of each copy of the molecule but also stores the
        molecule itself with the positions and orientations

        Args:
            atoms (object): The atom data
            positions (list): A list of positions
            orientations (list): A list of orientations
            name (str): The name of the molecule

        """

        # Convert to numpy arrays
        positions = np.array(positions, dtype="float32")
        orientations = np.array(orientations, dtype="float32")
        assert positions.shape == orientations.shape
        assert len(positions.shape) == 2
        assert positions.shape[1] == 3

        # Get the reference coords
        reference_coords = atoms.data[["x", "y", "z"]].to_numpy()

        # Loop through all the positions and rotations
        for position, rotation, orientation in zip(
            positions, Rotation.from_rotvec(orientations), orientations
        ):

            # Make a copy and apply the rotation and translation
            coords = (rotation.apply(reference_coords) + position).astype("float32")
            temp = atoms.data.copy()
            temp["x"] = coords[:, 0]
            temp["y"] = coords[:, 1]
            temp["z"] = coords[:, 2]

            # Add to the atoms
            self.add_atoms(AtomData(data=temp))

        # Add the molecule to it's own field
        if name is not None:
            self.__handle.sample.molecules[name].atoms = atoms.data
            self.__handle.sample.molecules[name].positions = positions
            self.__handle.sample.molecules[name].orientations = orientations

    @property
    def containing_box(self):
        """
        The box containing the sample

        """
        return self.__handle.sample.containing_box

    @containing_box.setter
    def containing_box(self, box):
        """
        The box containing the sample

        """
        self.__handle.sample.containing_box = box

    @property
    def centre(self):
        """
        The centre of the shape

        """
        return self.__handle.sample.centre

    @centre.setter
    def centre(self, centre):
        """
        The centre of the shape

        """
        self.__handle.sample.centre = centre

    @property
    def shape(self):
        """
        The ideal shape of the sample

        """
        return self.__handle.sample.shape

    @shape.setter
    def shape(self, shape):
        """
        The ideal shape of the sample

        """
        self.__handle.sample.shape = shape

    @property
    def shape_box(self):
        """
        Return the shape box

        """
        return shape_bounding_box(self.centre, self.shape)

    @property
    def shape_radius(self):
        """
        Return a radius

        """
        return shape_bounding_cylinder(self.centre, self.shape)[2]

    @property
    def bounding_box(self):
        """
        The bounding box

        Returns:
            tuple: lower, upper - the bounding box coordinates

        """
        return self.__handle.sample.bounding_box

    @bounding_box.setter
    def bounding_box(self, b):
        """
        The bounding box

        Returns:
            tuple: lower, upper - the bounding box coordinates

        """
        self.__handle.sample.bounding_box = b

    @property
    def molecules(self):
        """
        The list of molecule names

        Returns:
            list: The list of molecule names

        """
        return list(self.__handle.sample.molecules.keys())

    @property
    def number_of_molecular_models(self):
        """
        The number of molecular models

        Returns:
            int: The number of molecular models

        """
        return sum(
            molecule.positions.shape[0]
            for molecule in self.__handle.sample.molecules.values()
        )

    @property
    def number_of_molecules(self):
        """
        The number of molecules

        Returns:
            int: The number of molcules

        """
        return len(self.__handle.sample.molecules)

    def get_molecule(self, name):
        """
        Get the molcule data

        Args:
            name (str): The molecule name

        Returns:
            tuple (atoms, positions, orientations): The molecule data

        """
        return (
            AtomData(data=self.__handle.sample.molecules[name].atoms),
            self.__handle.sample.molecules[name].positions,
            self.__handle.sample.molecules[name].orientations,
        )

    def iter_molecules(self):
        """
        Iterate over the molecules

        Yields:
            tuple (name, data): The molecule data

        """
        for name in self.molecules:
            yield name, self.get_molecule(name)

    @property
    def number_of_atoms(self):
        """
        Returns:
            int: The number of atoms

        """
        return self.__handle.sample.atoms.number_of_atoms()

    def iter_atom_groups(self):
        """
        Iterate over the atom groups

        Yields:
            object: The atom group

        """
        return self.__handle.sample.atoms.bins()

    def iter_atoms(self):
        """
        Iterate over the atoms

        Yields:
            object: The atom data

        """
        for name, x in self.__handle.sample.atoms.bins():
            yield (x, x + self.step), AtomData(
                data=self.__handle.sample.atoms[name].atoms
            )

    def get_atoms(self):
        """
        Get all the atoms

        Returns:
            object: The atom data

        """

        # Iterate over dataset ranges
        data = pandas.DataFrame()
        for (x0, x1), atoms in self.iter_atoms():
            data = pandas.concat([data, atoms.data], ignore_index=True)
        return AtomData(data=data)

    def get_atoms_in_group(self, name):
        """
        Get the atoms in the group

        """
        return AtomData(data=self.__handle.sample.atoms[name].atoms)

    def get_atoms_in_range(self, x0, x1, filter=None):
        """
        Get the subset of atoms within the field of view

        Args:
            x0 (float): The start of the z range
            x1 (float): The end of the z range
            filter (func): A function to filter the atoms

        Returns:
            object: The atom data

        """

        # Cast input
        x0 = np.array(x0)
        x1 = np.array(x1)

        # Check input
        assert (x0 < x1).all()

        # Iterate over dataset ranges
        data = pandas.DataFrame()
        for x_min, x_max in self.atoms_dataset_range(x0, x1):
            name = self.atoms_dataset_name(x_min)
            atoms = self.__handle.sample.atoms[name].atoms
            coords = atoms[["x", "y", "z"]]
            atoms = atoms[((coords >= x0) & (coords < x1)).all(axis=1)]
            if filter is not None:
                atoms = filter(atoms)
            data = pandas.concat([data, atoms], ignore_index=True)
        return AtomData(data=data)

    def get_atoms_in_fov(self, x0, x1):
        """
        Get the subset of atoms within the field of view

        Args:
            x0 (float): The start of the z range
            x1 (float): The end of the z range

        Returns:
            object: The atom data

        """

        # Cast input
        x0 = np.array(x0)
        x1 = np.array(x1)

        # Check input
        assert (x0 < x1).all()

        # Iterate over dataset ranges
        atoms = self.get_atoms().data
        if len(atoms) > 0:
            coords = atoms[["x", "y"]]
            atoms = atoms[((coords >= x0) & (coords < x1)).all(axis=1)]
        return AtomData(data=atoms)

    def info(self):
        """
        Get some sample info

        Returns:
            str: Some sample info

        """
        lines = [
            "Sample information:",
            "    # Molecules:   %d" % self.number_of_molecules,
            "    # models:      %d" % self.number_of_molecular_models,
            "    # atoms:       %d" % self.number_of_atoms,
            "    Min x:         %.2f" % self.bounding_box[0][0],
            "    Min y:         %.2f" % self.bounding_box[0][1],
            "    Min z:         %.2f" % self.bounding_box[0][2],
            "    Max x:         %.2f" % self.bounding_box[1][0],
            "    Max y:         %.2f" % self.bounding_box[1][1],
            "    Max z:         %.2f" % self.bounding_box[1][2],
            "    Sample size x: %.2f" % self.dimensions[0],
            "    Sample size y: %.2f" % self.dimensions[1],
            "    Sample size z: %.2f" % self.dimensions[2],
            "    Min box x:     %.2f" % self.containing_box[0][0],
            "    Min box y:     %.2f" % self.containing_box[0][1],
            "    Min box z:     %.2f" % self.containing_box[0][2],
            "    Max box x:     %.2f" % self.containing_box[1][0],
            "    Max box y:     %.2f" % self.containing_box[1][1],
            "    Max box z:     %.2f" % self.containing_box[1][2],
            "    Centre x:      %.2f" % self.centre[0],
            "    Centre y:      %.2f" % self.centre[1],
            "    Centre z:      %.2f" % self.centre[2],
            "    Shape:         %s" % str(self.shape),
        ]
        for name, molecule in self.iter_molecules():
            molecule_lines = [
                "    Molecule: %s" % name,
                "        Positions:\n%s"
                % "\n".join(["%s%s" % (" " * 12, str(p)) for p in molecule[1]]),
                "        Orientations:\n%s"
                % "\n".join(["%s%s" % (" " * 12, str(o)) for o in molecule[2]]),
            ]
            lines.extend(molecule_lines)

        return "\n".join(lines)


class AtomSliceExtractor(object):
    """
    A class to select atoms given a rotation and translation

    """

    class Slice(object):
        """
        A class to keep the slice result

        """

        def __init__(self, data, x_min, x_max):
            self.atoms = data
            self.x_min = x_min
            self.x_max = x_max

    def __init__(self, sample, translation, rotation, x0, x1, thickness=10):
        """
        Preprocess the slices

        The sample will be translated along the X axis and then rotated. All
        atoms within the range x0 -> x1 will then be extracted in a number of
        slices of given thickness

        Args:
            sample (object): The sample object
            translation (float): The translation along X
            rotation (float): The rotation around the X axis
            x0 (array): The (x, y) minimum coordinate
            x1 (array): The (x, y) maximum coordinate
            thickness (float): The slice thickness in A

        """

        # Check if the rectangles overlaps with each other
        def overlapping(a0, a1, b0, b1):
            if not (
                a0[0] > b1[0]
                or a1[0] < b0[0]
                or a0[1] > b1[1]
                or a1[1] < b0[1]
                or a0[2] > b1[2]
                or a1[2] < b0[2]
            ):
                return True
            return False

        # Set the items
        self.sample = sample
        self.centre = sample.centre
        self.translation = translation
        self.rotation = rotation * pi / 180.0
        self.x0 = x0
        self.x1 = x1
        self.thickness = thickness

        # Loop through the atom groups
        group_coords = []
        group_names = []
        for name, xmin in self.sample.iter_atom_groups():
            xmax = xmin + self.sample.step

            # The coordinates of the corners of the group
            y = np.array(
                [
                    (xmin[0], xmin[1], xmin[2]),
                    (xmin[0], xmin[1], xmax[2]),
                    (xmin[0], xmax[1], xmin[2]),
                    (xmin[0], xmax[1], xmax[2]),
                    (xmax[0], xmin[1], xmin[2]),
                    (xmax[0], xmin[1], xmax[2]),
                    (xmax[0], xmax[1], xmin[2]),
                    (xmax[0], xmax[1], xmax[2]),
                ]
            )

            # Rotate the group and translate the corners of the group and
            # append to the list of groupd
            group_names.append(name)
            group_coords.append(
                Rotation.from_rotvec((self.rotation, 0, 0)).apply(y - self.centre)
                + self.centre
                - (self.translation, 0, 0)
            )

        # Compute the min and max coordinates of all the rotated groups and
        # compute the number of slices we will have
        group_coords = np.array(group_coords)
        min_x = np.min(group_coords.reshape((-1, 3)), axis=0)
        max_x = np.max(group_coords.reshape((-1, 3)), axis=0)
        self.min_z = max(0, min_x[2])
        self.max_z = max_x[2]
        num_slices = ceil((self.max_z - self.min_z) / self.thickness)

        # Loop through the groups. Compute the minimum and maximum coordinates
        # of the group and create a new box. Loop through the slices and check
        # if the box overlaps with the slice box. If it does then add the name
        # of the group to the list of groups in that slice.
        self.__groups_in_slice = defaultdict(list)
        for name, group in zip(group_names, group_coords):
            min_x = np.min(group, axis=0)
            max_x = np.max(group, axis=0)
            for i in range(num_slices):
                z0 = self.min_z + i * thickness
                z1 = self.min_z + (i + 1) * thickness
                if overlapping(min_x, max_x, (x0[0], x0[1], z0), (x1[0], x1[1], z1)):
                    self.__groups_in_slice[i].append(name)

    def __getitem__(self, index):
        """
        Get the slice of atom data

        Args:
            index (int): The slice index

        Returns:
            object: The slice

        """

        # Get the min and max coordinate of the slice
        z0 = self.min_z + index * self.thickness
        z1 = self.min_z + (index + 1) * self.thickness
        x_min = (self.x0[0], self.x0[1], z0)
        x_max = (self.x1[0], self.x1[1], z1)

        # Rotate and translate the atoms and select only those atoms within the
        # slice
        def filter_atoms(atoms):
            coords = atoms[["x", "y", "z"]].to_numpy()
            coords = (
                Rotation.from_rotvec((0, self.rotation, 0)).apply(coords - self.centre)
                + self.centre
                - (0, self.translation, 0)
            ).astype("float32")
            atoms["x"] = coords[:, 0]
            atoms["y"] = coords[:, 1]
            atoms["z"] = coords[:, 2]
            atoms = atoms[((coords >= x_min) & (coords < x_max)).all(axis=1)]
            return atoms

        # Check the index
        assert index >= 0 and index < len(self.__groups_in_slice)

        # Loop through the ranges and extract the data
        data = []
        for name in self.__groups_in_slice[index]:
            atoms = filter_atoms(self.sample.get_atoms_in_group(name).data)
            if len(atoms) > 0:
                data.append(atoms)

        # Cat data
        if len(data) > 0:
            data = pandas.concat(data)
        else:
            data = pandas.DataFrame()

        # Return the slice
        return AtomSliceExtractor.Slice(AtomData(data=data), x_min, x_max)

    def __len__(self):
        """
        Returns:
            int: The number of slices

        """
        return len(self.__groups_in_slice)

    def __iter__(self):
        """
        Iterate through the slices

        """
        for i in range(len(self)):
            item = self[i]
            if len(item.atoms.data) > 0:
                yield self[i]


class AtomDeleter(object):
    """
    A class to delete atoms so we can place a molecule

    Create a grid with the molecule atom positions. Then test the atom
    positions in the sample to and only select those atoms which don't fall on
    molecule atom positions.

    """

    def __init__(self, atoms, position=(0, 0, 0), rotation=(0, 0, 0)):
        """
        Create the grid

        Args:
            atoms (object): the molcule atoms
            position (array): The molecule position
            rotation (array): The molecule rotation

        """

        # Compute the transformed coordinates
        coords = (
            Rotation.from_rotvec(rotation).apply(atoms.data[["x", "y", "z"]]) + position
        )

        # The min and max coordinates
        border = 1  # A
        self.x0 = np.floor(np.min(coords, axis=0)) - border
        self.x1 = np.ceil(np.max(coords, axis=0)) + border

        # The size of the box in A
        self.size = self.x1 - self.x0

        # The grid size
        self.grid_cell_size = 1.0  # 2 A
        min_distance = (
            2.0 / self.grid_cell_size
        )  # The distance at which to delete atoms

        # The dimensions of the grid
        grid_shape = np.ceil(self.size / self.grid_cell_size).astype("uint32")

        # Allocate the grid
        self.grid = np.ones(shape=grid_shape, dtype="bool")
        # self.grid = ~self.grid
        # X, Y, Z = np.mgrid[0:grid_shape[0],0:grid_shape[1],0:grid_shape[2]]
        # R = np.sqrt((X-grid_shape[0]/2)**2+(Y-grid_shape[1]/2)**2+(Z-grid_shape[2]/2)**2)
        # r = abs(self.x0[0] - self.x1[0])
        # print(R.max(), grid_shape[0]/2.0)
        # self.grid = R < (grid_shape[0]/2.0)
        # return

        # Get the indices and set the grid values. The fill in any holes using
        # a morphological fill. The distance transform finds the distance to
        # any zero holes so in this case we need to set the nodes with protein
        # to False then compute the distance and then set any nodes within a
        # given distance to True. So the mask is True where we want to delete
        # atoms.
        indices = np.floor((coords - self.x0) / self.grid_cell_size).astype("int32")
        index_x = indices[:, 0]
        index_y = indices[:, 1]
        index_z = indices[:, 2]
        assert (indices >= 0).all()
        self.grid[index_x, index_y, index_z] = False
        self.grid = ~self.grid
        # from matplotlib import pylab
        # pylab.imshow(self.grid[:,:,grid_shape[2]//2])
        # pylab.show()

        self.grid = (
            scipy.ndimage.morphology.distance_transform_edt(self.grid) < min_distance
        )
        return
        # self.grid = scipy.ndimage.morphology.binary_closing(self.grid, iterations=2)
        # self.grid = scipy.ndimage.morphology.binary_fill_holes(self.grid)
        # import mrcfile

        # outfile = mrcfile.new("temp.mrc", overwrite=True)
        # outfile.set_data(self.grid.astype("int8"))
        # logger.info("Filled grid with atom positions:")
        # logger.info("    x0: %g" % self.x0[0])
        # logger.info("    y0: %g" % self.x0[1])
        # logger.info("    z0: %g" % self.x0[2])
        # logger.info("    x1: %g" % self.x1[0])
        # logger.info("    y1: %g" % self.x1[1])
        # logger.info("    z1: %g" % self.x1[2])
        # logger.info("    num elements: %d" % self.grid.size)
        # logger.info("    num filled: %d" % np.count_nonzero(self.grid))

    def __call__(self, atoms):
        """
        Select atoms that don't fall on the molecule atom positions

        Args:
            atoms (object): The atoms to select from

        Returns:
            object: The selected atoms

        """

        # Get the atom coords
        coords = atoms[["x", "y", "z"]]

        # Compute the indices
        indices = (
            np.floor((coords - self.x0) / self.grid_cell_size)
            .astype("int32")
            .to_numpy()
        )
        # The coordinates inside the grid
        inside_grid = ((indices >= 0) & (indices < self.grid.shape)).all(axis=1)

        # The selection
        selection = np.zeros(shape=coords.shape[0], dtype="bool")
        selection[~inside_grid] = True

        # Compute the selection
        indices_inside_grid = indices[inside_grid]
        index_x = indices_inside_grid[:, 0]
        index_y = indices_inside_grid[:, 1]
        index_z = indices_inside_grid[:, 2]
        selection[inside_grid] = self.grid[index_x, index_y, index_z] == False

        # Print some info
        logger.info(
            "Deleted %d/%d atoms"
            % (len(selection) - np.count_nonzero(selection), coords.shape[0])
        )

        # Return the atoms
        return atoms[selection]


def load(filename, mode="r"):
    """
    Load the sample

    Args:
        filename (str): The filename of the sample
        mode (str): The opening mode

    Returns:
        object: The test sample

    """
    return Sample(filename, mode=mode)


def add_ice(sample, centre=None, shape=None, density=940.0, pack=False):
    """
    Create a sample with just a load of water molecules

    Args:
        sample (object): The sample object
        shape (object): The shape description
        density (float): The water density

    Returns:
        object: The sample object

    """

    def shape_filter_coordinates(coords, centre, shape):
        def cube_filter_coordinates(coords, centre, cube):
            length = cube["length"]
            x0 = centre - length / 2.0
            x1 = centre + length / 2.0
            return coords[((coords >= x0) & (coords < x1)).all(axis=1)]

        def cuboid_filter_coordinates(coords, centre, cuboid):
            length_x = cuboid["length_x"]
            length_y = cuboid["length_y"]
            length_z = cuboid["length_z"]
            length = np.array((length_x, length_y, length_z))
            x0 = centre - length / 2.0
            x1 = centre + length / 2.0
            return coords[((coords >= x0) & (coords < x1)).all(axis=1)]

        def cylinder_filter_coordinates(coords, centre, cylinder):
            length = cylinder["length"]
            radius = cylinder["radius"]
            y0 = centre - length / 2.0
            y1 = centre + length / 2.0
            x = coords[:, 0]
            y = coords[:, 1]
            z = coords[:, 2]
            return coords[
                (y >= y0[0])
                & (y < y1[0])
                & ((z - centre[2]) ** 2 + (x - centre[1]) ** 2 <= radius**2)
            ]

        # Filter the coords
        return {
            "cube": cube_filter_coordinates,
            "cuboid": cuboid_filter_coordinates,
            "cylinder": cylinder_filter_coordinates,
        }[shape["type"]](coords, centre, shape[shape["type"]])

    # Cast input
    centre = np.array(centre)

    # Get the filename of the water.cif file
    filename = parakeet.data.get_path("water.cif")

    # Create a single ribosome sample
    single_water = AtomData.from_ligand_file(filename)
    atoms = single_water.data[0:3]
    water_coords = atoms[["x", "y", "z"]].copy()
    water_coords -= water_coords.iloc[0].copy()

    # Set the volume in A^3
    if shape["type"] == "cube":
        length = shape["cube"]["length"]
        volume = length**3
        length_x = length
        length_y = length
        length_z = length
    elif shape["type"] == "cuboid":
        length_x = shape["cuboid"]["length_x"]
        length_y = shape["cuboid"]["length_y"]
        length_z = shape["cuboid"]["length_z"]
        volume = length_x * length_y * length_z
    elif shape["type"] == "cylinder":
        length = shape["cylinder"]["length"]
        radius = shape["cylinder"]["radius"]
        volume = pi * radius**2 * length
        length_x = 2 * radius
        length_y = length
        length_z = 2 * radius
    else:
        raise RuntimeError("Unknown shape")

    # Get the centre and offset
    centre_x, centre_y, centre_z = centre
    offset_x = centre_x - length_x / 2.0
    offset_y = centre_y - length_y / 2.0
    offset_z = centre_z - length_z / 2.0
    offset = np.array((offset_x, offset_y, offset_z), dtype="float32")

    # Determine the number of waters to place
    avogadros_number = scipy.constants.Avogadro
    molar_mass_of_water = 18.01528  # grams / mole
    density_of_water = density  # kg / m^3
    mass_of_water = (density_of_water * 1000) * (volume * 1e-10**3)  # g
    number_of_waters = int(
        floor((mass_of_water / molar_mass_of_water) * avogadros_number)
    )

    # Uniform random or packed
    if not pack:

        # The water filename
        filename = parakeet.data.get_path("water.cif")

        # Get the water coords
        single_water = parakeet.sample.AtomData.from_ligand_file(filename)
        atoms = single_water.data[0:3]
        water_coords = atoms[["x", "y", "z"]].copy()

        # Min = 0
        water_coords -= water_coords.min()

        # Translation
        if shape["type"] != "cylinder":
            x = np.random.uniform(offset_x, offset_x + length_x, size=number_of_waters)
            y = np.random.uniform(offset_y, offset_y + length_y, size=number_of_waters)
            z = np.random.uniform(offset_z, offset_z + length_z, size=number_of_waters)
        else:
            r = radius * np.sqrt(np.random.uniform(0, 1, size=number_of_waters))
            t = np.random.uniform(0, 2 * pi, size=number_of_waters)
            x = centre_x + r * np.cos(t)
            y = np.random.uniform(offset_y, offset_y + length_z, size=number_of_waters)
            z = centre_z + r * np.sin(t)
        translation = np.array((x, y, z)).T

        # Random orientations
        rotation = Rotation.from_rotvec(random_uniform_rotation(number_of_waters))

        # Rotate the Hydrogens around the Oxygen and translate
        O = rotation.apply(water_coords.iloc[0].copy()) + translation
        H1 = rotation.apply(water_coords.iloc[1].copy()) + translation
        H2 = rotation.apply(water_coords.iloc[2].copy()) + translation

        def create_atom_data(atomic_number, coords):

            # Create a new array
            def new_array(size, name, value):
                return (
                    np.zeros(
                        shape=(size,), dtype=parakeet.sample.AtomData.column_data[name]
                    )
                    + value
                )

            # Create the new arrays
            atomic_number = new_array(coords.shape[0], "atomic_number", atomic_number)
            sigma = new_array(coords.shape[0], "sigma", 0.085)
            occupancy = new_array(coords.shape[0], "occupancy", 1.0)
            charge = new_array(coords.shape[0], "charge", 0)

            # Return the data frame
            return pandas.DataFrame(
                {
                    "atomic_number": atomic_number,
                    "x": coords[:, 0],
                    "y": coords[:, 1],
                    "z": coords[:, 2],
                    "sigma": sigma,
                    "occupancy": occupancy,
                    "charge": charge,
                }
            )

        # Add the sample atoms
        data_buffer = []
        data_buffer.append(create_atom_data(8, O))
        data_buffer.append(create_atom_data(1, H1))
        data_buffer.append(create_atom_data(1, H2))

        sample.add_atoms(AtomData(data=pandas.concat(data_buffer, ignore_index=True)))

    else:

        # Van der Waals radius of water
        van_der_waals_radius = 2.7 / 2.0  # A

        # Compute the total volume in the spheres
        volume_of_spheres = (
            (4.0 / 3.0) * pi * van_der_waals_radius**3 * number_of_waters
        )

        # Create the grid. The sphere packer takes the grid is (z, y, x) but
        # because it does slices along Z and we want slices along X we flip the X
        # and Z grid spec here
        grid = (
            int(floor(length_x / (2 * van_der_waals_radius))),
            int(floor(length_y / (2 * van_der_waals_radius))),
            int(floor(length_z / (2 * van_der_waals_radius))),
        )

        # Compute the node length and density
        node_length = max((length_z / grid[2], length_y / grid[1], length_x / grid[0]))
        sphere_density = number_of_waters / volume

        logger.info(
            f"Initialising Sphere Packer:\n"
            f"    Length X:           {length_x} A\n"
            f"    Length Y:           {length_y} A\n"
            f"    Length Z:           {length_z} A\n"
            f"    Volume:             {volume} A^3\n"
            f"    Density of water:   {density_of_water} Kg/m^3\n"
            f"    Mass of water:      {mass_of_water} g\n"
            f"    Number of waters:   {number_of_waters}\n"
            f"    Mean diameter:      {2*van_der_waals_radius} A\n"
            f"    Volume filled:      {100*volume_of_spheres/volume} %\n"
            f"    Packer grid:        ({grid[0]}, {grid[1]}, {grid[2]})\n"
            f"    Grid node length:   {node_length} A\n"
            f"    Density of spheres: {sphere_density} #/A^3\n"
        )

        # Create the sphere packer
        packer = parakeet.freeze.SpherePacker(
            grid, node_length, sphere_density, van_der_waals_radius, max_iter=10
        )

        # Extract all the data
        logger.info("Generating water positions:")
        start_time = time.time()
        max_buffer = 10_000_000
        data_buffer = []
        for x_index, x_slice in enumerate(packer):

            # Read the coordinates. The packer goes along the z axis so we need to
            # flip the coordinates since we want x slices
            coords = []
            for node in x_slice:
                coords.extend(node)
            coords = np.flip(np.array(coords, dtype="float32"), axis=1) + offset

            # Filter the coordinates by the shape to ensure no ice is outside the
            # shape. This is only really necessary for the cylinder shape
            coords = shape_filter_coordinates(coords, centre, shape)

            # Compute the rotation
            rotation = Rotation.from_rotvec(random_uniform_rotation(coords.shape[0]))

            # Rotate the Hydrogens around the Oxygen and translate
            O = rotation.apply(water_coords.iloc[0].copy()) + coords
            H1 = rotation.apply(water_coords.iloc[1].copy()) + coords
            H2 = rotation.apply(water_coords.iloc[2].copy()) + coords

            def create_atom_data(atomic_number, coords):

                # Create a new array
                def new_array(size, name, value):
                    return (
                        np.zeros(shape=(size,), dtype=AtomData.column_data[name])
                        + value
                    )

                # Create the new arrays
                atomic_number = new_array(
                    coords.shape[0], "atomic_number", atomic_number
                )
                sigma = new_array(coords.shape[0], "sigma", 0.085)
                occupancy = new_array(coords.shape[0], "occupancy", 1.0)
                charge = new_array(coords.shape[0], "charge", 0)

                # Return the data frame
                return pandas.DataFrame(
                    {
                        "atomic_number": atomic_number,
                        "x": coords[:, 0],
                        "y": coords[:, 1],
                        "z": coords[:, 2],
                        "sigma": sigma,
                        "occupancy": occupancy,
                        "charge": charge,
                    }
                )

            # Add the sample atoms
            data_buffer.append(create_atom_data(8, O))
            data_buffer.append(create_atom_data(1, H1))
            data_buffer.append(create_atom_data(1, H2))
            if sum(b.shape[0] for b in data_buffer) > max_buffer:
                logger.info(
                    "    Writing %d atoms" % sum(b.shape[0] for b in data_buffer)
                )
                sample.add_atoms(
                    AtomData(data=pandas.concat(data_buffer, ignore_index=True))
                )
                data_buffer = []

            # The estimates time left
            time_taken = time.time() - start_time
            estimated_time = (len(packer) - x_index) * time_taken / (x_index + 1)

            # Log some info
            logger.info(
                "    X slice %d/%d: Num molecules: %d (remaining %.d seconds)"
                % (x_index, len(packer), O.shape[0], estimated_time)
            )

        # Add anything remaining in the data buffer
        if len(data_buffer) > 0:
            logger.info("    Writing %d atoms" % sum(b.shape[0] for b in data_buffer))
            sample.add_atoms(
                AtomData(data=pandas.concat(data_buffer, ignore_index=True))
            )
            del data_buffer

        # Print some output
        logger.info(f"Sphere packer: Num unplaced:  {packer.num_unplaced_samples()}")

    # Return the sample
    return sample


def new(filename, box=None, centre=None, shape=None, ice=None, coords=None, **kwargs):
    """
    Create the sample

    Args:
        filename (str): The filename of the sample
        box (list): The containing box of the sample
        centre (list): The centering of the sample in the box
        shape (object): The shape of the sample
        ice (object): The ice description
        coords_filename (object): The coordinates

    Returns:
        object: The test sample

    """

    # Check the dimensions are valid
    assert is_shape_inside_box(box, centre, shape)

    # Create the sample
    sample = Sample(filename, mode="w")

    # Set the sample box and shape
    sample.containing_box = ((0, 0, 0), box)
    sample.centre = centre
    sample.shape = shape

    # Add some ice
    if ice is not None and ice["generate"]:
        add_ice(sample, centre, shape, ice["density"])

    # Add atoms from coordinates file
    if coords is not None and coords["filename"]:
        atoms = AtomData.from_gemmi_file(coords["filename"])
        if coords["recentre"]:
            atoms.data = recentre(atoms.data)
            position = sample.centre
        else:
            position = (0, 0, 0)

        # Set the orientation
        orientation = coords["orientation"]
        if orientation is None or len(orientation) != 3:
            orientation = random_uniform_rotation()[0]

        # Add the molecule
        sample.add_molecule(
            atoms, positions=[position], orientations=[orientation], name=None
        )

    # Print some info
    logger.info(sample.info())

    # Get the sample
    return sample


def add_single_molecule(sample, name):
    """
    Create a sample with a single molecule

    The molecule will be positioned at the centre

    Args:
        sample (object): The sample object
        name (str): The name of the molecule to add

    Returns:
        object: The sample

    """
    logger.info("Adding single %s molecule" % name)

    # Get the filename of the 4v5d.cif file
    filename = parakeet.data.get_pdb(name)

    # Get the atom data
    atoms = AtomData.from_gemmi_file(filename)
    atoms.data = recentre(atoms.data)

    # Get atom data bounds
    coords = atoms.data[["x", "y", "z"]]
    x0 = np.min(coords, axis=0) + sample.centre
    x1 = np.max(coords, axis=0) + sample.centre

    # Check the coords
    assert is_box_inside_shape((x0, x1), sample.centre, sample.shape)

    # Print some info
    logger.info(
        "\n".join(
            (
                "Name:   %s" % name,
                "Min x:  %.2f" % x0[0],
                "Min y:  %.2f" % x0[1],
                "Min z:  %.2f" % x0[2],
                "Max x:  %.2f" % x1[0],
                "Max y:  %.2f" % x1[1],
                "Max z:  %.2f" % x1[2],
                "Size x: %.2f" % (x1[0] - x0[0]),
                "Size y: %.2f" % (x1[1] - x0[1]),
                "Size z: %.2f" % (x1[2] - x0[2]),
                "Pos x:  %.2f" % (sample.centre[0]),
                "Pos y:  %.2f" % (sample.centre[1]),
                "Pos z:  %.2f" % (sample.centre[2]),
            )
        )
    )

    position = sample.centre

    # Delete the atoms where we want to place the molecules
    sample.del_atoms(AtomDeleter(atoms, position, (0, 0, 0)))

    # Add the molecule
    sample.add_molecule(
        atoms, positions=[position], orientations=[(0, 0, 0)], name=name
    )

    # Return the sample
    return sample


def add_multiple_molecules(sample, molecules):
    """
    Create a sample with multiple molecules

    The molecules will be positioned randomly in the sample

    Args:
        sample (object): The sample object
        molecules (object): The molecules

    Returns:
        object: The sample

    """

    # Setup some arrays
    atom_data = {}
    all_labels = []
    all_boxes = []
    all_positions = []
    all_orientations = []

    # Generate the orientations and boxes
    for name, items in molecules.items():

        # Skip if number is zero
        if len(items) == 0:
            continue

        # Print some info
        logger.info("Adding %d %s molecules" % (len(items), name))

        # Get the filename of the 4v5d.cif file
        filename = parakeet.data.get_pdb(name)

        # Get the atom data
        atoms = AtomData.from_gemmi_file(filename)
        atoms.data = recentre(atoms.data)
        atom_data[name] = atoms

        # Generate some random orientations
        logger.info("    Generating random orientations")
        for item in items:
            rotation = item.get("orientation", None)
            if rotation is None:
                rotation = random_uniform_rotation(1)[0]
            rotation = Rotation.from_rotvec(rotation)
            coords = rotation.apply(atoms.data[["x", "y", "z"]])
            all_orientations.append(rotation.as_rotvec())
            all_positions.append(item.get("position", None))
            all_boxes.append(np.max(coords, axis=0) - np.min(coords, axis=0))
            all_labels.append(name)

    # Put the molecules in the sample
    logger.info("Placing molecules:")
    if any(p is None or len(p) == 0 for p in all_positions):
        all_positions = distribute_boxes_uniformly(
            shape_enclosed_box(sample.centre, sample.shape), all_boxes
        )

    # Set the positions and orientations by molecule
    positions = defaultdict(list)
    orientations = defaultdict(list)
    for label, rotation, position, box in zip(
        all_labels, all_orientations, all_positions, all_boxes
    ):

        # Get atom data bounds
        x0 = position - box / 2.0
        x1 = position + box / 2.0

        # Check the coords
        assert is_box_inside_shape((x0, x1), sample.centre, sample.shape)

        # Construct arrays
        positions[label].append(position)
        orientations[label].append(rotation)

        # Delete the atoms where we want to place the molecules
        sample.del_atoms(AtomDeleter(atom_data[label], position, rotation))

        # Print some info
        logger.info(
            "    placing %s at (%.2f, %.2f, %.2f) with orientation (%.2f, %.2f, %.2f)"
            % ((label,) + tuple(position) + tuple(rotation))
        )

    # Add the molecules
    for name in atom_data.keys():
        sample.add_molecule(
            atom_data[name],
            positions=positions[name],
            orientations=orientations[name],
            name=name,
        )


def add_molecules(filename, molecules=None, **kwargs):
    """
    Take a sample and add a load of molecules

    Args:
        filename (str): The filename of the sample
        molecules (dict): The molecules

    Returns:
        object: The test sample

    """

    # Convert to list of positions/orientations
    temp = {}
    for key, value in molecules.items():
        if isinstance(value, int):
            temp[key] = [{} for i in range(value)]
        else:
            temp[key] = value
    molecules = temp

    # The total number of molecules
    total_number_of_molecules = sum(map(len, molecules.values()))

    # Open the sample
    sample = Sample(filename, mode="r+")

    # Put the molecule in the centre if only one
    if total_number_of_molecules == 0:
        raise RuntimeError("Need at least 1 molecule")
    elif total_number_of_molecules == 1:
        key = [key for key, value in molecules.items() if len(value) > 0][0]
        if molecules[key][0].get("position", None) is None:
            molecules[key][0]["position"] = sample.centre
        if molecules[key][0].get("orientation", None) is None:
            molecules[key][0]["orientation"] = (0, 0, 0)

    # Add the molecules
    add_multiple_molecules(sample, molecules)

    # Show some info
    logger.info(sample.info())

    # Return the sample
    return sample


def mill(filename, box=None, centre=None, shape=None, **kwargs):
    """
    Mill the sample

    """

    def shape_filter_coordinates(coords, centre, shape):
        def cube_filter_coordinates(coords, centre, cube):
            length = cube["length"]
            x0 = centre - length / 2.0
            x1 = centre + length / 2.0
            return ((coords >= x0) & (coords < x1)).all(axis=1)

        def cuboid_filter_coordinates(coords, centre, cuboid):
            length_x = cuboid["length_x"]
            length_y = cuboid["length_y"]
            length_z = cuboid["length_z"]
            length = np.array((length_x, length_y, length_z))
            x0 = centre - length / 2.0
            x1 = centre + length / 2.0
            return ((coords >= x0) & (coords < x1)).all(axis=1)

        def cylinder_filter_coordinates(coords, centre, cylinder):
            length = cylinder["length"]
            radius = cylinder["radius"]
            y0 = centre - length / 2.0
            y1 = centre + length / 2.0
            x = coords["x"]
            y = coords["y"]
            z = coords["z"]
            return (
                (y >= y0[0])
                & (y < y1[0])
                & ((z - centre[2]) ** 2 + (x - centre[1]) ** 2 <= radius**2)
            )

        # Filter the coords
        return {
            "cube": cube_filter_coordinates,
            "cuboid": cuboid_filter_coordinates,
            "cylinder": cylinder_filter_coordinates,
        }[shape["type"]](coords, centre, shape[shape["type"]])

    class Deleter(object):
        def __init__(self, sample):
            self.sample = sample
            self.x0, self.x1 = self.sample.bounding_box

        def __call__(self, atoms):
            selection = shape_filter_coordinates(
                atoms[["x", "y", "z"]], self.sample.centre, self.sample.shape
            )
            return atoms[selection]

    # Open the sample
    sample = Sample(filename, mode="r+")

    # Set the sample box and shape
    sample.containing_box = ((0, 0, 0), box)
    sample.centre = centre
    sample.shape = shape

    # Delete the atoms
    sample.del_atoms(Deleter(sample))

    atoms = sample.get_atoms()
    coords = atoms.data[["x", "y", "z"]]
    x0 = coords.min()
    x1 = coords.max()
    sample.bounding_box = (x0, x1)

    # Print some info
    logger.info(sample.info())


def sputter(filename, element=None, thickness=20):
    """
    Add a sputter coating to the sample of the desired thickness

    This is very crude and adds atoms at random positions

    Params:
        atomic_number (int): The atomic number
        density (float): The density of atoms (g/cm3)
        thickness (float): The thickness (A)


    """

    # Open the sample
    sample = Sample(filename, mode="r+")

    # Get the sample shape
    shape = sample.shape
    centre = sample.centre

    # Set the volume in A^3
    if shape["type"] == "cube":
        length = shape["cube"]["length"]
        length_x = length
        length_y = length
        length_z = length
        sputter_length_x = length_x + thickness * 2
        sputter_length_y = length_y + thickness * 2
        sputter_length_z = length_z + thickness * 2
        shape_volume = length**3
        # sputter_volume = (
        #     sputter_length_x * sputter_length_y * sputter_length_z - shape_volume
        # )
    elif shape["type"] == "cuboid":
        length_x = shape["cuboid"]["length_x"]
        length_y = shape["cuboid"]["length_y"]
        length_z = shape["cuboid"]["length_z"]
        sputter_length_x = length_x + thickness * 2
        sputter_length_y = length_y + thickness * 2
        sputter_length_z = length_z + thickness * 2
        shape_volume = length_x * length_y * length_z
        # sputter_volume = (
        #     sputter_length_x * sputter_length_y * sputter_length_z - shape_volume
        # )
    elif shape["type"] == "cylinder":
        length = shape["cylinder"]["length"]
        radius = shape["cylinder"]["radius"]
        length_x = 2 * radius
        length_y = length
        length_z = 2 * radius
        sputter_length = length + thickness * 2
        sputter_radius = radius + thickness * 2
        shape_volume = pi * radius**2 * length
        # sputter_volume = pi * sputter_radius ** 2 * sputter_length - shape_volume
    else:
        raise RuntimeError("Unknown shape")

    # Determine the number of atoms to place
    if element is None:
        return
    elif element == "C":
        molar_mass = 12.0107  # grams / mole
        density = 2.3
        atomic_number = 6
    elif element == "Ir":
        molar_mass = 192.217  # grams / mole
        density = 22.56
        atomic_number = 77
    elif element == "Cr":
        molar_mass = 51.9961  # grams / mole
        density = 7.15
        atomic_number = 24
    elif element == "Pt":
        molar_mass = 195.084  # grams / mole
        density = 21.45
        atomic_number = 78
    else:
        raise RuntimeError("Unknown element %s" % element)
    avogadros_number = scipy.constants.Avogadro
    number_density = ((density * 1e-8**3) / molar_mass) * avogadros_number
    print("Placing %g %s atoms per A^3" % (number_density, element))

    # Get the centre and offset
    centre_x, centre_y, centre_z = centre
    offset_x = centre_x - length_x / 2.0
    offset_y = centre_y - length_y / 2.0
    offset_z = centre_z - length_z / 2.0
    # offset = np.array((offset_x, offset_y, offset_z), dtype="float32")

    # Translation
    if shape["type"] != "cylinder":
        # Generate positions in the thickness range
        volume = length_x * length_y * thickness
        number_of_atoms = int(number_density * volume)
        x = np.random.uniform(offset_x, offset_x + length_x, size=number_of_atoms)
        y = np.random.uniform(offset_y, offset_y + length_y, size=number_of_atoms)
        z = np.random.uniform(offset_z - thickness, offset_z, size=number_of_atoms)
        print("Placed %d atoms" % number_of_atoms)
        print("WARNING - ONLY APPLYING TO TOP OF SAMPLE")
    else:
        raise RuntimeError("Not implemented")
    position = np.array((x, y, z)).T

    def create_atom_data(atomic_number, coords):

        # Create a new array
        def new_array(size, name, value):
            return (
                np.zeros(
                    shape=(size,), dtype=parakeet.sample.AtomData.column_data[name]
                )
                + value
            )

        # Create the new arrays
        atomic_number = new_array(coords.shape[0], "atomic_number", atomic_number)
        sigma = new_array(coords.shape[0], "sigma", 0.085)
        occupancy = new_array(coords.shape[0], "occupancy", 1.0)
        charge = new_array(coords.shape[0], "charge", 0)

        # Return the data frame
        return pandas.DataFrame(
            {
                "atomic_number": atomic_number,
                "x": coords[:, 0],
                "y": coords[:, 1],
                "z": coords[:, 2],
                "sigma": sigma,
                "occupancy": occupancy,
                "charge": charge,
            }
        )

    # Add the sample atoms
    data_buffer = []
    data_buffer.append(create_atom_data(atomic_number, position))

    sample.add_atoms(AtomData(data=pandas.concat(data_buffer, ignore_index=True)))



#
# parakeet.beam.py
#
# Copyright (C) 2019 Diamond Light Source and Rosalind Franklin Institute
#
# Author: James Parkhurst
#
# This code is distributed under the GPLv3 license, a copy of
# which is included in the root directory of this package.
#


class Beam(object):
    """
    A class to encapsulate a beam

    """

    def __init__(
        self,
        energy=300,
        energy_spread=0,
        acceleration_voltage_spread=0,
        source_spread=0.1,
        electrons_per_angstrom=30,
        theta=0,
        phi=0,
        drift=None,
        defocus_drift=None,
    ):
        """
        Initialise the beam

        Args:
            energy (float): The beam energy (keV)
            energy_spread (float): dE / E where dE is the 1/e half width
            acceleration_voltage_spread (float): dV / V where dV is the 1 / e half width
            source_spread (float): The source spread (mrad)
            electrons_per_angstrom (float): The number of electrons per angstrom
            theta (float): The beam tilt theta
            phi (float): The beam tilt phi
            drift (obj): The beam drift
            defocus_drift (obj): The defocus drift

        """
        self.energy = energy
        self.energy_spread = energy_spread
        self.acceleration_voltage_spread = acceleration_voltage_spread
        self.source_spread = source_spread
        self.electrons_per_angstrom = electrons_per_angstrom
        self.theta = theta
        self.phi = phi
        self.drift = drift
        self.defocus_drift = defocus_drift


def new(
    energy=None,
    energy_spread=None,
    acceleration_voltage_spread=None,
    source_spread=0.1,
    electrons_per_angstrom=None,
    theta=0,
    phi=0,
    drift=None,
    defocus_drift=None,
):
    """
    Create a beam

    Args:
        energy (float): The beam energy (keV)
        energy_spread (float): dE / E where dE is the 1/e half width
        acceleration_voltage_spread (float): dV / V where dV is the 1 / e half width
        source_spread (float): The source spread (mrad)
        electrons_per_angstrom (float): The number of electrons per angstrom
        theta (float): The beam tilt theta
        phi (float): The beam tilt phi
        drift (obj): The beam drift
        defocus_drift (obj): The defocus drift

    Returns:
        object: The beam object

    """
    return Beam(
        energy=energy,
        energy_spread=energy_spread,
        acceleration_voltage_spread=acceleration_voltage_spread,
        source_spread=source_spread,
        electrons_per_angstrom=electrons_per_angstrom,
        theta=theta,
        phi=phi,
        drift=drift,
        defocus_drift=defocus_drift,
    )



#
# parakeet.command_line.simulate.py
#
# Copyright (C) 2019 Diamond Light Source and Rosalind Franklin Institute
#
# Author: James Parkhurst
#
# This code is distributed under the GPLv3 license, a copy of
# which is included in the root directory of this package.
#
import argparse
import logging
import numpy
import time
import parakeet.io
import parakeet.command_line
import parakeet.config
import parakeet.microscope
import parakeet.sample
import parakeet.scan
import parakeet.simulation
from math import pi

# Get the logger
logger = logging.getLogger(__name__)


def projected_potential(args=None):
    """
    Simulate the projected potential from the sample

    """

    # Get the start time
    start_time = time.time()

    # Create the argument parser
    parser = argparse.ArgumentParser(
        description="Simulate the exit wave from the sample"
    )

    # Add some command line arguments
    parser.add_argument(
        "-c",
        "--config",
        type=str,
        default=None,
        dest="config",
        help="The yaml file to configure the simulation",
    )
    parser.add_argument(
        "-d",
        "--device",
        choices=["cpu", "gpu"],
        default=None,
        dest="device",
        help="Choose the device to use",
    )
    parser.add_argument(
        "--cluster.max_workers",
        type=int,
        default=None,
        dest="cluster_max_workers",
        help="The maximum number of worker processes",
    )
    parser.add_argument(
        "--cluster.method",
        type=str,
        choices=["sge"],
        default=None,
        dest="cluster_method",
        help="The cluster method to use",
    )
    parser.add_argument(
        "-s",
        "--sample",
        type=str,
        default="sample.h5",
        dest="sample",
        help="The filename for the sample",
    )

    # Parse the arguments
    args = parser.parse_args(args=args)

    # Configure some basic logging
    parakeet.command_line.configure_logging()

    # Set the command line args in a dict
    command_line = {}
    if args.device is not None:
        command_line["device"] = args.device
    if args.cluster_max_workers is not None or args.cluster_method is not None:
        command_line["cluster"] = {}
    if args.cluster_max_workers is not None:
        command_line["cluster"]["max_workers"] = args.cluster_max_workers
    if args.cluster_method is not None:
        command_line["cluster"]["method"] = args.cluster_method

    # Load the full configuration
    config = parakeet.config.load(args.config, command_line)

    # Print some options
    parakeet.config.show(config)

    # Create the microscope
    microscope = parakeet.microscope.new(**config["microscope"])

    # Create the sample
    logger.info(f"Loading sample from {args.sample}")
    sample = parakeet.sample.load(args.sample)

    # Create the scan
    if config["scan"]["step_pos"] == "auto":
        radius = sample.shape_radius
        config["scan"]["step_pos"] = config["scan"]["step_angle"] * radius * pi / 180.0
    scan = parakeet.scan.new(**config["scan"])
    if scan.positions[-1] > sample.containing_box[1][0]:
        raise RuntimeError("Scan goes beyond sample containing box")

    # Create the simulation
    simulation = parakeet.simulation.projected_potential(
        microscope=microscope,
        sample=sample,
        scan=scan,
        device=config["device"],
        simulation=config["simulation"],
        cluster=config["cluster"],
    )

    # Run the simulation
    simulation.run()

    # Write some timing stats
    logger.info("Time taken: %.2f seconds" % (time.time() - start_time))


def exit_wave(args=None):
    """
    Simulate the exit wave from the sample

    """

    # Get the start time
    start_time = time.time()

    # Create the argument parser
    parser = argparse.ArgumentParser(
        description="Simulate the exit wave from the sample"
    )

    # Add some command line arguments
    parser.add_argument(
        "-c",
        "--config",
        type=str,
        default=None,
        dest="config",
        help="The yaml file to configure the simulation",
    )
    parser.add_argument(
        "-d",
        "--device",
        choices=["cpu", "gpu"],
        default=None,
        dest="device",
        help="Choose the device to use",
    )
    parser.add_argument(
        "--cluster.max_workers",
        type=int,
        default=None,
        dest="cluster_max_workers",
        help="The maximum number of worker processes",
    )
    parser.add_argument(
        "--cluster.method",
        type=str,
        choices=["sge"],
        default=None,
        dest="cluster_method",
        help="The cluster method to use",
    )
    parser.add_argument(
        "-s",
        "--sample",
        type=str,
        default="sample.h5",
        dest="sample",
        help="The filename for the sample",
    )
    parser.add_argument(
        "-e",
        "--exit_wave",
        type=str,
        default="exit_wave.h5",
        dest="exit_wave",
        help="The filename for the exit wave",
    )

    # Parse the arguments
    args = parser.parse_args(args=args)

    # Configure some basic logging
    parakeet.command_line.configure_logging()

    # Set the command line args in a dict
    command_line = {}
    if args.device is not None:
        command_line["device"] = args.device
    if args.cluster_max_workers is not None or args.cluster_method is not None:
        command_line["cluster"] = {}
    if args.cluster_max_workers is not None:
        command_line["cluster"]["max_workers"] = args.cluster_max_workers
    if args.cluster_method is not None:
        command_line["cluster"]["method"] = args.cluster_method

    # Load the full configuration
    config = parakeet.config.load(args.config, command_line)

    # Print some options
    parakeet.config.show(config)

    # Create the microscope
    microscope = parakeet.microscope.new(**config["microscope"])

    # Create the sample
    logger.info(f"Loading sample from {args.sample}")
    sample = parakeet.sample.load(args.sample)

    # Create the scan
    if config["scan"]["step_pos"] == "auto":
        radius = sample.shape_radius
        config["scan"]["step_pos"] = config["scan"]["step_angle"] * radius * pi / 180.0
    scan = parakeet.scan.new(**config["scan"])

    # Create the simulation
    simulation = parakeet.simulation.exit_wave(
        microscope=microscope,
        sample=sample,
        scan=scan,
        device=config["device"],
        simulation=config["simulation"],
        cluster=config["cluster"],
    )

    # Create the writer
    logger.info(f"Opening file: {args.exit_wave}")
    writer = parakeet.io.new(
        args.exit_wave,
        shape=simulation.shape,
        pixel_size=simulation.pixel_size,
        dtype=numpy.complex64,
    )

    # Run the simulation
    simulation.run(writer)

    # Write some timing stats
    logger.info("Time taken: %.2f seconds" % (time.time() - start_time))


def optics(args=None):
    """
    Simulate the optics

    """

    # Get the start time
    start_time = time.time()

    # Create the argument parser
    parser = argparse.ArgumentParser(description="Simulate the optics")

    # Add some command line arguments
    parser.add_argument(
        "-c",
        "--config",
        type=str,
        default=None,
        dest="config",
        help="The yaml file to configure the simulation",
    )
    parser.add_argument(
        "-d",
        "--device",
        choices=["cpu", "gpu"],
        default=None,
        dest="device",
        help="Choose the device to use",
    )
    parser.add_argument(
        "--cluster.max_workers",
        type=int,
        default=None,
        dest="cluster_max_workers",
        help="The maximum number of worker processes",
    )
    parser.add_argument(
        "--cluster.method",
        type=str,
        choices=["sge"],
        default=None,
        dest="cluster_method",
        help="The cluster method to use",
    )
    parser.add_argument(
        "-e",
        "--exit_wave",
        type=str,
        default="exit_wave.h5",
        dest="exit_wave",
        help="The filename for the exit wave",
    )
    parser.add_argument(
        "-o",
        "--optics",
        type=str,
        default="optics.h5",
        dest="optics",
        help="The filename for the optics",
    )

    # Parse the arguments
    args = parser.parse_args(args=args)

    # Configure some basic logging
    parakeet.command_line.configure_logging()

    # Set the command line args in a dict
    command_line = {}
    if args.device is not None:
        command_line["device"] = args.device
    if args.cluster_max_workers is not None or args.cluster_method is not None:
        command_line["cluster"] = {}
    if args.cluster_max_workers is not None:
        command_line["cluster"]["max_workers"] = args.cluster_max_workers
    if args.cluster_method is not None:
        command_line["cluster"]["method"] = args.cluster_method

    # Load the full configuration
    config = parakeet.config.load(args.config, command_line)

    # Print some options
    parakeet.config.show(config)

    # Create the microscope
    microscope = parakeet.microscope.new(**config["microscope"])

    # Create the exit wave data
    logger.info(f"Loading sample from {args.exit_wave}")
    exit_wave = parakeet.io.open(args.exit_wave)

    # Create the scan
    config["scan"]["angles"] = exit_wave.angle
    config["scan"]["positions"] = exit_wave.position[:, 1]
    scan = parakeet.scan.new(**config["scan"])

    # Create the simulation
    simulation = parakeet.simulation.optics(
        microscope=microscope,
        exit_wave=exit_wave,
        scan=scan,
        device=config["device"],
        simulation=config["simulation"],
        sample=config["sample"],
        cluster=config["cluster"],
    )

    # Create the writer
    logger.info(f"Opening file: {args.optics}")
    writer = parakeet.io.new(
        args.optics,
        shape=simulation.shape,
        pixel_size=simulation.pixel_size,
        dtype=numpy.float32,
    )

    # Run the simulation
    simulation.run(writer)

    # Write some timing stats
    logger.info("Time taken: %.2f seconds" % (time.time() - start_time))


def ctf(args=None):
    """
    Simulate the ctf

    """

    # Get the start time
    start_time = time.time()

    # Create the argument parser
    parser = argparse.ArgumentParser(description="Simulate the ctf")

    # Add some command line arguments
    parser.add_argument(
        "-c",
        "--config",
        type=str,
        default=None,
        dest="config",
        help="The yaml file to configure the simulation",
    )
    parser.add_argument(
        "-o",
        type=str,
        default="ctf.h5",
        dest="output",
        help="The filename for the output",
    )

    # Parse the arguments
    args = parser.parse_args(args=args)

    # Configure some basic logging
    parakeet.command_line.configure_logging()

    # Load the full configuration
    config = parakeet.config.load(args.config)

    # Print some options
    parakeet.config.show(config)

    # Create the microscope
    microscope = parakeet.microscope.new(**config["microscope"])

    # Create the simulation
    simulation = parakeet.simulation.ctf(
        microscope=microscope, simulation=config["simulation"]
    )

    # Create the writer
    logger.info(f"Opening file: {args.output}")
    writer = parakeet.io.new(
        args.output,
        shape=simulation.shape,
        pixel_size=simulation.pixel_size,
        dtype=numpy.complex64,
    )

    # Run the simulation
    simulation.run(writer)

    # Write some timing stats
    logger.info("Time taken: %.2f seconds" % (time.time() - start_time))


def image(args=None):
    """
    Simulate the image with noise

    """

    # Get the start time
    start_time = time.time()

    # Create the argument parser
    parser = argparse.ArgumentParser(description="Simulate the image")

    # Add some command line arguments
    parser.add_argument(
        "-c",
        "--config",
        type=str,
        default=None,
        dest="config",
        help="The yaml file to configure the simulation",
    )
    parser.add_argument(
        "-o",
        "--optics",
        type=str,
        default="optics.h5",
        dest="optics",
        help="The filename for the optics",
    )
    parser.add_argument(
        "-i",
        "--image",
        type=str,
        default="image.h5",
        dest="image",
        help="The filename for the image",
    )

    # Parse the arguments
    args = parser.parse_args(args=args)

    # Configure some basic logging
    parakeet.command_line.configure_logging()

    # Load the full configuration
    config = parakeet.config.load(args.config)

    # Print some options
    parakeet.config.show(config)

    # Create the microscope
    microscope = parakeet.microscope.new(**config["microscope"])

    # Create the exit wave data
    logger.info(f"Loading sample from {args.optics}")
    optics = parakeet.io.open(args.optics)

    # Create the scan
    config["scan"]["angles"] = optics.angle
    config["scan"]["positions"] = optics.position[:, 1]
    scan = parakeet.scan.new(**config["scan"])

    # Create the simulation
    simulation = parakeet.simulation.image(
        microscope=microscope,
        optics=optics,
        scan=scan,
        device=config["device"],
        simulation=config["simulation"],
        cluster=config["cluster"],
    )

    # Create the writer
    logger.info(f"Opening file: {args.image}")
    writer = parakeet.io.new(
        args.image,
        shape=simulation.shape,
        pixel_size=simulation.pixel_size,
        dtype=numpy.float32,
    )

    # Run the simulation
    simulation.run(writer)

    # Write some timing stats
    logger.info("Time taken: %.2f seconds" % (time.time() - start_time))


def simple():
    """
    Simulate the image

    """

    # Get the start time
    start_time = time.time()

    # Create the argument parser
    parser = argparse.ArgumentParser(description="Simulate the image")

    # Add some command line arguments
    parser.add_argument(
        "-c",
        "--config",
        type=str,
        default=None,
        dest="config",
        help="The yaml file to configure the simulation",
    )
    parser.add_argument(
        "-o",
        "--output",
        type=str,
        default="output.h5",
        dest="output",
        help="The filename for the output",
    )
    parser.add_argument(
        "atoms",
        type=str,
        default=None,
        nargs="?",
        help="The filename for the input atoms",
    )

    # Parse the arguments
    args = parser.parse_args()

    # Configure some basic logging
    parakeet.command_line.configure_logging()

    # Load the full configuration
    config = parakeet.config.load(args.config)

    # Print some options
    parakeet.config.show(config)

    # Create the microscope
    microscope = parakeet.microscope.new(**config["microscope"])

    # Create the exit wave data
    logger.info(f"Loading sample from {args.atoms}")
    atoms = parakeet.sample.AtomData.from_text_file(args.atoms)

    # Create the simulation
    simulation = parakeet.simulation.simple(
        microscope=microscope,
        atoms=atoms,
        device=config["device"],
        simulation=config["simulation"],
    )

    # Create the writer
    logger.info(f"Opening file: {args.output}")
    writer = parakeet.io.new(
        args.output,
        shape=simulation.shape,
        pixel_size=simulation.pixel_size,
        dtype=numpy.complex64,
    )

    # Run the simulation
    simulation.run(writer)

    # Write some timing stats
    logger.info("Time taken: %.2f seconds" % (time.time() - start_time))



#
# parakeet.lens.py
#
# Copyright (C) 2019 Diamond Light Source and Rosalind Franklin Institute
#
# Author: James Parkhurst
#
# This code is distributed under the GPLv3 license, a copy of
# which is included in the root directory of this package.
#


class Lens(object):
    """
    A class to encapsulate a lens

    """

    def __init__(
        self,
        m=0,
        c_10=20,
        c_12=0.0,
        phi_12=0.0,
        c_21=0.0,
        phi_21=0.0,
        c_23=0.0,
        phi_23=0.0,
        c_30=0.04,
        c_32=0.0,
        phi_32=0.0,
        c_34=0.0,
        phi_34=0.0,
        c_41=0.0,
        phi_41=0.0,
        c_43=0.0,
        phi_43=0.0,
        c_45=0.0,
        phi_45=0.0,
        c_50=0.0,
        c_52=0.0,
        phi_52=0.0,
        c_54=0.0,
        phi_54=0.0,
        c_56=0.0,
        phi_56=0.0,
        inner_aper_ang=0.0,
        outer_aper_ang=0.0,
        c_c=0.0,
        current_spread=0.0,
    ):
        """
        Initialise the lens

        Args:
            m (int): The vortex momentum
            c_10 (float): The defocus (A)
            c_12 (float): The 2-fold astigmatism (A)
            phi_12 (float): The azimuthal angle of 2-fold astigmatism (degrees)
            c_21 (float): The axial coma (A)
            phi_21 (float): The azimuthal angle of axial coma (degrees)
            c_23 (float): The 3-fold astigmatism (A)
            phi_23 (float): The azimuthal angle of 3-fold astigmatism (degrees)
            c_30 (float): The 3rd order spherical aberration (mm)
            c_32 (float): The axial star aberration (A)
            phi_32 (float): The azimuthal angle of axial star aberration (degrees)
            c_34 (float): The 4-fold astigmatism (A)
            phi_34 (float): The azimuthal angle of 4-fold astigmatism (degrees)
            c_41 (float): The 4th order axial coma (A)
            phi_41 (float): The azimuthal angle of 4th order axial coma (degrees)
            c_43 (float): The 3-lobe aberration (A)
            phi_43 (float): The azimuthal angle of 3-lobe aberration (degrees)
            c_45 (float): The 5-fold astigmatism (A)
            phi_45 (float): The azimuthal angle of 5-fold astigmatism (degrees)
            c_50 (float): The 5th order spherical aberration (mm)
            c_52 (float): The 5th order axial star aberration (A)
            phi_52 (float): The azimuthal angle of 5th order axial star aberration (degrees)
            c_54 (float): The 5th order rosette aberration (A)
            phi_54 (float): The azimuthal angle of 5th order rosette aberration (degrees)
            c_56 (float): The 6-fold astigmatism (A)
            phi_56 (float): The azimuthal angle of 6-fold astigmatism (degrees)
            inner_aper_ang (float): The inner aperture (mrad)
            outer_aper_ang (float): The outer aperture (mrad)
            c_c (float): The chromatic abberation (mm)
            current_spread (float): dI / I where dI is the 1/e half width

        """
        self.m = m
        self.c_10 = c_10
        self.c_12 = c_12
        self.phi_12 = phi_12
        self.c_21 = c_21
        self.phi_21 = phi_21
        self.c_23 = c_23
        self.phi_23 = phi_23
        self.c_30 = c_30
        self.c_32 = c_32
        self.phi_32 = phi_32
        self.c_34 = c_34
        self.phi_34 = phi_34
        self.c_41 = c_41
        self.phi_41 = phi_41
        self.c_43 = c_43
        self.phi_43 = phi_43
        self.c_45 = c_45
        self.phi_45 = phi_45
        self.c_50 = c_50
        self.c_52 = c_52
        self.phi_52 = phi_52
        self.c_54 = c_54
        self.phi_54 = phi_54
        self.c_56 = c_56
        self.phi_56 = phi_56
        self.inner_aper_ang = inner_aper_ang
        self.outer_aper_ang = outer_aper_ang
        self.c_c = c_c
        self.current_spread = current_spread


def new(**kwargs):
    """
    Create a lens

    Returns:
        object: The lens object

    """
    return Lens(**kwargs)



#
# parakeet.freeze.__init__.py
#
# Copyright (C) 2019 Diamond Light Source and Rosalind Franklin Institute
#
# Author: James Parkhurst
#
# This code is distributed under the GPLv3 license, a copy of
# which is included in the root directory of this package.
#
import logging
import numpy
import pandas
import scipy.ndimage.morphology
from math import ceil, floor

from parakeet_ext import *  # noqa

# Get the logger
logger = logging.getLogger(__name__)


def freeze(atoms, x0, x1):
    """
    Given the input set of atoms add water molecules to random positions
    to simulate the sample in vitreous ice.

    Args:
        atoms (object): The list of atom positions
        x0 (array): The starting coordinate of the bounding box
        x1 (array): The ending coordinate of the bounding box

    Returns:
        object: The new list of atom positions with water molecules

    """

    # The grid size
    grid_size = 10.0  # 10 A

    # Get the min and max atom positions
    x = atoms["x"]
    y = atoms["y"]
    z = atoms["z"]
    min_x = numpy.min(x)
    max_x = numpy.max(x)
    min_y = numpy.min(y)
    max_y = numpy.max(y)
    min_z = numpy.min(z)
    max_z = numpy.max(z)

    # Ensure the atoms are within the box
    assert min_x >= x0[0]
    assert min_y >= x0[1]
    assert min_z >= x0[2]
    assert max_x <= x1[0]
    assert max_y <= x1[1]
    assert max_z <= x1[2]

    # Create a grid at 10A spacing to determine the shape of the input molecule
    shape = (
        int(ceil((x1[2] - x0[2]) / grid_size)),
        int(ceil((x1[1] - x0[1]) / grid_size)),
        int(ceil((x1[0] - x0[0]) / grid_size)),
    )
    logger.info("Allocating grid")
    logger.info("    shape: %s" % str(shape))
    grid = numpy.zeros(shape=shape, dtype="bool")

    # Fill in any spaces in the grid
    x_index = numpy.floor((x - x0[0]) / grid_size).astype("int32")
    y_index = numpy.floor((y - x0[1]) / grid_size).astype("int32")
    z_index = numpy.floor((z - x0[2]) / grid_size).astype("int32")
    grid[z_index, y_index, x_index] = True
    grid = scipy.ndimage.morphology.binary_fill_holes(grid)
    logger.info("Filled grid with atom positions:")
    logger.info("    x0: %g" % x0[0])
    logger.info("    y0: %g" % x0[1])
    logger.info("    z0: %g" % x0[2])
    logger.info("    x1: %g" % x1[0])
    logger.info("    y1: %g" % x1[1])
    logger.info("    z1: %g" % x1[2])
    logger.info("    num elements: %d" % grid.size)
    logger.info("    num filled: %d" % numpy.sum(grid))

    # Determine the remaining volume
    x_length = x1[0] - x0[0]
    y_length = x1[1] - x0[1]
    z_length = x1[2] - x0[2]
    total_volume = x_length * y_length * z_length  # A^3
    filled_volume = numpy.sum(grid) * 10**3
    remaining_volume = total_volume - filled_volume

    # Determine the number of waters to place
    avogadros_number = scipy.constants.Avogadro
    molar_mass_of_water = 18.01528  # grams / mole
    density_of_water = 997  # kg / m^3
    mass_of_water = (density_of_water * 1000) * (remaining_volume * 1e-10**3)  # g
    number_of_waters = int(
        floor((mass_of_water / molar_mass_of_water) * avogadros_number)
    )

    # Print some stuff of water
    logger.info("Water information:")
    logger.info("    Total volume %g m^3" % (total_volume * 1e-10**3))
    logger.info("    Sample volume %g m^3" % (filled_volume * 1e-10**3))
    logger.info("    Volume of water: %g m^3" % (remaining_volume * 1e-10**3))
    logger.info("    Density of water: %g kg/m^3" % density_of_water)
    logger.info("    Mass of water: %g kg" % (mass_of_water / 1000))
    logger.info("    Number of water molecules to place: %d" % number_of_waters)

    # Loop adding waters until complete
    logger.info("Placing waters:")
    number_to_place = number_of_waters
    water_coords = numpy.zeros(shape=(0, 3), dtype="float64")
    while number_to_place > 0:
        coords = numpy.random.uniform(x0, x1, size=(number_to_place, 3))
        x_index = numpy.floor((coords[:, 0] - x0[0]) / grid_size).astype("int32")
        y_index = numpy.floor((coords[:, 1] - x0[1]) / grid_size).astype("int32")
        z_index = numpy.floor((coords[:, 2] - x0[2]) / grid_size).astype("int32")
        selection = grid[z_index, y_index, x_index] == False
        coords = coords[selection, :]
        water_coords = numpy.concatenate((water_coords, coords), axis=0)
        number_to_place -= len(coords)
        logger.info("    placed %d waters" % len(coords))

    # Create the data frame for the atoms
    # FIXME Currently only adding oxygens since they are major contribution
    # to scattering of electrons; however, we should also add hydrogens
    logger.info("Creating water atom data")
    shape = number_of_waters
    water_atoms = pandas.DataFrame(
        {
            "model": numpy.zeros(shape=shape, dtype="uint32"),
            "chain": numpy.zeros(shape=shape, dtype="str"),
            "residue": numpy.zeros(shape=shape, dtype="str"),
            "atomic_number": numpy.full(shape=shape, fill_value=8, dtype="uint32"),
            "x": water_coords[:, 0],
            "y": water_coords[:, 1],
            "z": water_coords[:, 2],
            "occ": numpy.ones(shape=shape, dtype="float64"),
            "charge": numpy.zeros(shape=shape, dtype="uint32"),
            "sigma": numpy.full(shape=shape, fill_value=0.085, dtype="float64"),
            "region": numpy.zeros(shape=shape, dtype="uint32"),
        }
    )

    # Return the atom positions
    return pandas.concat([atoms, water_atoms])



#
# Copyright (C) 2019 Diamond Light Source and Rosalind Franklin Institute
#
# Author: James Parkhurst
#
# This code is distributed under the GPLv3 license, a copy of
# which is included in the root directory of this package.
#
import numpy as np
import parakeet.landau
import scipy.signal
from math import sqrt, pi, cos, exp, log, ceil, floor


def effective_thickness(shape, angle):
    """
    Compute the effective thickness

    """
    TINY = 1e-10
    if shape["type"] == "cube":
        D0 = shape["cube"]["length"]
        thickness = D0 / (cos(pi * angle / 180.0) + TINY)
    elif shape["type"] == "cuboid":
        D0 = shape["cuboid"]["length_z"]
        thickness = D0 / (cos(pi * angle / 180.0) + TINY)
    elif shape["type"] == "cylinder":
        thickness = shape["cylinder"]["radius"] * 2
    return thickness


def zero_loss_fraction(shape, angle):
    """
    Compute the zero loss fraction

    """
    thickness = effective_thickness(shape, angle)
    mean_free_path = 3150  # A for Amorphous Ice at 300 keV
    electron_fraction = exp(-thickness / mean_free_path)
    return electron_fraction


def mp_loss_fraction(shape, angle):
    """
    Compute the inelastic fraction

    """
    thickness = effective_thickness(shape, angle)
    mean_free_path = 3150  # A for Amorphous Ice at 300 keV
    electron_fraction = exp(-thickness / mean_free_path)
    return 1.0 - electron_fraction


def fraction_of_electrons(shape, angle, model=None):
    """
    Compute the fraction of electrons

    """
    if model is None:
        fraction = 1.0
    elif model == "zero_loss":
        fraction = zero_loss_fraction(shape, angle)
    elif model == "mp_loss":
        fraction = mp_loss_fraction(shape, angle)
    elif model == "unfiltered":
        fraction = 1.0
    elif model == "cc_corrected":
        fraction = 1.0
    return fraction


def most_probable_loss(energy, shape, angle):
    """
    Compute the MPL peak and sigma

    Params:
        energy (float): Beam energy in keV

    Returns:
        tuple: (peak, sigma) of the energy loss distribution (eV)

    """
    thickness = effective_thickness(shape, angle)
    thickness = min(thickness, 100000)  # Maximum 10 um - to avoid issues at high tilt
    peak, fwhm = parakeet.landau.mpl_and_fwhm(energy, thickness)
    return peak, fwhm / (2 * sqrt(2 * log(2)))


class EnergyFilterOptimizer(object):
    """
    A simple class to find the optimal energy filter placement when trying to
    take into account the inelastic scattering component

    """

    def __init__(self, energy_spread=0.8, dE_min=-10, dE_max=200, dE_step=0.01):
        """
        Initialise the class

        Params:
            energy_spread (float): The energy spread (eV)
            dE_min (float): The minimum energy loss (eV)
            dE_max (float): The maximum energy loss (eV)
            dE_step (float): The energy loss step (eV)

        """

        # Save the energy spread
        self.energy_spread = energy_spread  # eV

        # Save an instance of the landau distribution class
        self.landau = parakeet.landau.Landau()

        # Set the energy losses to consider
        self.dE_min = dE_min
        self.dE_max = dE_max
        self.dE_step = dE_step

    def __call__(self, energy, thickness, filter_width=None):
        """
        Compute the optimum position for the energy filter

        Params:
            energy (float): The beam energy (eV)
            thickness (float): The sample thickness (A)
            filter_width (float): The energy filter width (eV)

        Returns:
            float: The optimal position to maximum number of electrons (eV)

        """

        # Check the input
        assert energy > 0
        assert thickness > 0

        # The energy loss distribution
        dE, distribution = self.energy_loss_distribution(energy, thickness)

        # Compute optimum given the position and filter width
        if filter_width is None:
            position = np.sum(dE * distribution) / np.sum(dE)
        else:
            size = len(distribution)
            kernel_size = filter_width / self.dE_step
            ks = kernel_size / 2
            kx = np.arange(-size // 2, size // 2 + 1)
            kernel = np.exp(-0.5 * (kx / ks) ** 80)
            kernel = kernel / np.sum(kernel)
            num = scipy.signal.fftconvolve(distribution, kernel, mode="same")
            position = dE[np.argmax(num)]

        # Return the filter position
        return position

    def elastic_fraction(self, energy, thickness):
        """
        Compute the elastic electron fraction

        Params:
            energy (float): The beam energy (eV)
            thickness (float): The sample thickness (A)

        Returns:
            float: The elastic electron fraction

        """
        # Compute the fractions for the zero loss and energy losses
        mean_free_path = 3150  # A for Amorphous Ice at 300 keV
        return exp(-thickness / mean_free_path)

    def energy_loss_distribution(self, energy, thickness):
        """
        Compute the energy loss distribution

        Params:
            energy (float): The beam energy (eV)
            thickness (float): The sample thickness (A)

        Returns:
            (array, array): dE (ev) and the energy loss distribution

        """

        # The energy losses to consider
        dE = np.arange(self.dE_min, self.dE_max, self.dE_step, dtype="float64")

        # The energy loss distribution
        energy_loss_distribution = self.landau(dE, energy, thickness)
        energy_loss_distribution /= self.dE_step * np.sum(energy_loss_distribution)

        # The zero loss distribution
        zero_loss_distribution = (1.0 / sqrt(pi * self.energy_spread**2)) * np.exp(
            -(dE**2) / self.energy_spread**2
        )

        # Compute the fractions for the zero loss and energy losses
        elastic_fraction = self.elastic_fraction(energy, thickness)

        # Return the distribution
        distribution = (
            elastic_fraction * zero_loss_distribution
            + (1 - elastic_fraction) * energy_loss_distribution
        )
        return dE, distribution

    def compute_elastic_component(self, energy, thickness, position, filter_width):
        """
        Compute the elastic fraction and energy spread

        Params:
            energy (float): The beam energy (eV)
            thickness (float): The sample thickness (A)
            position (float): The filter position (eV)
            filter_width (float): The energy filter width (eV)

        Returns:
            float, float: The electron fraction and energy spread (eV)

        """

        # The energy losses to consider
        dE = np.arange(self.dE_min, self.dE_max, self.dE_step)

        # The zero loss distribution
        P = (1.0 / sqrt(pi * self.energy_spread**2)) * np.exp(
            -(dE**2) / self.energy_spread**2
        )
        C = self.dE_step * np.cumsum(P)

        # Compute the fractions for the zero loss and energy losses
        fraction = self.elastic_fraction(energy, thickness)

        # Compute the number of electrons and the sigma
        if filter_width is not None:
            dE0 = position - filter_width / 2.0
            dE1 = position + filter_width / 2.0
            x0 = int(floor((dE0 - self.dE_min) / self.dE_step))
            x1 = int(ceil((dE1 - self.dE_min) / self.dE_step))
            x0 = max(x0, 0)
            x1 = min(x1, len(P) - 1)
            assert x1 > x0
            fraction *= C[x1] - C[x0]
            P = P[x0:x1]
            dE = dE[x0:x1]

        # Compute the spread
        if len(P) > 0 and np.sum(P) > 0:
            dE_mean = np.sum(P * dE) / np.sum(P)
            spread = sqrt(np.sum(P * (dE - dE_mean) ** 2) / np.sum(P)) * sqrt(2)
        else:
            spread = 0

        # Return the fraction and spread
        return fraction, spread

    def compute_inelastic_component(self, energy, thickness, position, filter_width):
        """
        Compute the inelastic fraction and energy spread

        Params:
            energy (float): The beam energy (eV)
            thickness (float): The sample thickness (A)
            position (float): The filter position (eV)
            filter_width (float): The energy filter width (eV)

        Returns:
            float, float: The electron fraction and energy spread (eV)

        """

        # The energy losses to consider
        dE = np.arange(self.dE_min, self.dE_max, self.dE_step)

        # The energy loss distribution
        P = self.landau(dE, energy, thickness)
        P /= self.dE_step * np.sum(P)
        C = np.cumsum(P) * self.dE_step

        # Compute the fractions for the zero loss and energy losses
        fraction = 1 - self.elastic_fraction(energy, thickness)

        # Compute the number of electrons and the sigma
        if filter_width is not None:
            dE0 = position - filter_width / 2.0
            dE1 = position + filter_width / 2.0
            x0 = int(floor((dE0 - self.dE_min) / self.dE_step))
            x1 = int(ceil((dE1 - self.dE_min) / self.dE_step))
            x0 = max(x0, 0)
            x1 = min(x1, len(P) - 1)
            assert x1 > x0
            fraction *= C[x1] - C[x0]
            P = P[x0:x1]
            dE = dE[x0:x1]

        # Compute the spread
        if len(P) > 0 and np.sum(P) > 0:
            peak, fwhm = parakeet.landau.mpl_and_fwhm(energy / 1000, thickness)
            sigma = fwhm / (2 * sqrt(2 * log(2)))

            # This is a hack to compute the variance because the heavy tail of
            # the landau distribution makes the variance undefined and this
            # causes problems for large filter sizes. There is probably a
            # better way to do this so maybe should look at fixing.
            P *= np.exp(-0.5 * (dE - peak) ** 2 / (2 * sigma) ** 2)

            P /= np.sum(P)
            dE_mean = np.sum(P * dE) / np.sum(P)
            spread = sqrt(np.sum(P * (dE - dE_mean) ** 2) / np.sum(P)) * sqrt(2)
        else:
            spread = 0

        # Return the fraction and spread
        return fraction, spread



#
# parakeet.command_line.config.py
#
# Copyright (C) 2019 Diamond Light Source and Rosalind Franklin Institute
#
# Author: James Parkhurst
#
# This code is distributed under the GPLv3 license, a copy of
# which is included in the root directory of this package.
#
import argparse
import logging
import yaml
import parakeet.config
import parakeet.command_line

# Get the logger
logger = logging.getLogger(__name__)


def show():
    """
    Show the full configuration

    """
    # Create the argument parser
    parser = argparse.ArgumentParser(description="Show the configuration")

    # Add some command line arguments
    parser.add_argument(
        "-c",
        "--config",
        type=str,
        default=None,
        dest="config",
        help="The yaml file to configure the simulation",
    )

    # Configure some basic logging
    parakeet.command_line.configure_logging()

    # Parse the arguments
    config = parakeet.config.load(parser.parse_args().config)

    # Print some options
    parakeet.config.show(config, full=True)


def edit():
    """
    Edit the configuration

    """
    # Create the argument parser
    parser = argparse.ArgumentParser(description="Edit the configuration")

    # Add some command line arguments
    parser.add_argument(
        "-i",
        "--input",
        type=str,
        default="config.yaml",
        dest="input",
        help="The input yaml file to configure the simulation",
    )

    parser.add_argument(
        "-o",
        "--output",
        type=str,
        default="config_new.yaml",
        dest="output",
        help="The output yaml file to configure the simulation",
    )

    parser.add_argument(
        "-s", type=str, default="", dest="config", help="The configuration string"
    )

    # Parse the command line
    args = parser.parse_args()

    # Configure some basic logging
    parakeet.command_line.configure_logging()

    # Parse the arguments
    config = parakeet.config.load(args.input, yaml.safe_load(args.config))

    # Print the config
    parakeet.config.show(config, full=True)

    # Save the config
    with open(args.output, "w") as outfile:
        yaml.safe_dump(config, outfile)



#
# Copyright (C) 2019 Diamond Light Source and Rosalind Franklin Institute
#
# Author: James Parkhurst
#
# This code is distributed under the GPLv3 license, a copy of
# which is included in the root directory of this package.
#
import parakeet.sample
import argparse
import multem
import numpy
import os.path
import pandas
import scipy.constants
import urllib.request
from math import ceil, pi, log
from matplotlib import pylab


def get_water_model_filename():
    """
    Get the water model filename

    """
    return "water_645_coords.pdb"


def get_water_atomic_model():
    """
    Download the water model from zenodo

    """
    filename = get_water_model_filename()
    if not os.path.exists(filename):
        print("Downloading %s" % filename)
        link = "https://zenodo.org/record/4415836/files/%s?download=1" % filename
        urllib.request.urlretrieve(link, filename)


def load_water_atomic_model():
    """
    Extract the water atomic model to a file

    """
    # Read the atom data
    if not os.path.exists("atoms.csv"):
        print("Reading model from %s" % get_water_model_filename())
        atom_data = parakeet.sample.AtomData.from_gemmi_file(get_water_model_filename())
        atom_data.data.to_csv("atoms.csv")
    else:
        print("Reading model from %s" % "atoms.csv")
        atom_data = parakeet.sample.AtomData(data=pandas.read_csv("atoms.csv"))
    return atom_data


def next_power_2(x):
    """
    Get the next power of two

    """
    return 2 ** (ceil(log(x, 2)))


def radial_average(data):
    """
    Compute the radial average

    """
    Y, X = numpy.indices((data.shape))
    ysize, xsize = data.shape
    distance = numpy.sqrt((X - xsize / 2.0) ** 2 + (Y - ysize / 2.0) ** 2)
    distance = numpy.floor(distance).astype("int32")
    c1 = numpy.bincount(distance.ravel()).astype("float32")
    c2 = numpy.bincount(distance.ravel(), data.ravel()).astype("float32")
    return c2 / c1


def create_input_multislice():
    """
    Create the input multislice parameters

    """

    # Initialise the input and system configuration
    input_multislice = multem.Input()

    # Set simulation experiment
    input_multislice.simulation_type = "EWRS"

    # Electron-Specimen interaction model
    input_multislice.interaction_model = "Multislice"
    # input_multislice.potential_type = "Lobato_0_12"
    input_multislice.potential_type = "Peng_0_12"

    # Potential slicing
    input_multislice.potential_slicing = "dz_Proj"

    # Electron-Phonon interaction model
    input_multislice.pn_model = "Still_Atom"

    # Specimen thickness
    input_multislice.thick_type = "Whole_Spec"

    # Illumination model
    input_multislice.illumination_model = "Partial_Coherent"
    input_multislice.temporal_spatial_incoh = "Temporal_Spatial"

    # Set the energy
    input_multislice.E_0 = 300

    return input_multislice


def compute_projected_potential():
    """
    Compute the projected potential in slices

    """

    def compute(atom_data, pixel_size, thickness):

        # Get the dimensions
        x_min = atom_data.data["x"].min()
        x_max = atom_data.data["x"].max()
        y_min = atom_data.data["y"].min()
        y_max = atom_data.data["y"].max()
        z_min = atom_data.data["z"].min()
        z_max = atom_data.data["z"].max()
        x_size = x_max - x_min
        y_size = y_max - y_min
        z_size = z_max - z_min

        # Translate to centre
        x_box_size = x_size
        y_box_size = y_size
        z_box_size = z_size
        x_trans = (x_box_size - x_size) / 2.0 - x_min
        y_trans = (y_box_size - y_size) / 2.0 - y_min
        z_trans = (z_box_size - z_size) / 2.0 - z_min
        atom_data.translate((x_trans, y_trans, z_trans))

        # Trim the atom data
        z_select_min = z_box_size / 2 - thickness / 2
        z_select_max = z_box_size / 2 + thickness / 2
        selection = (atom_data.data["z"] > z_select_min) & (
            atom_data.data["z"] < z_select_max
        )
        atom_data = parakeet.sample.AtomData(data=atom_data.data[selection])
        num_atoms = len(atom_data.data)

        # Create the system configuration
        system_conf = multem.SystemConfiguration()
        system_conf.precision = "float"
        system_conf.device = "device"

        # Create the input multislice configuration
        input_multislice = create_input_multislice()

        # Compute the number of pixels
        nx = int(x_box_size / pixel_size)
        ny = int(y_box_size / pixel_size)
        x_box_size = nx * pixel_size
        y_box_size = ny * pixel_size

        # Create the specimen size
        input_multislice.nx = nx
        input_multislice.ny = ny
        input_multislice.spec_lx = x_box_size
        input_multislice.spec_ly = y_box_size
        input_multislice.spec_lz = z_box_size
        input_multislice.spec_dz = thickness

        # Set the specimen atoms
        input_multislice.spec_atoms = atom_data.to_multem()

        potential = []

        def callback(z0, z1, V):
            V = numpy.array(V)
            potential.append(V)

        # Run the simulation
        multem.compute_projected_potential(system_conf, input_multislice, callback)

        # Save the potential
        potential = numpy.sum(potential, axis=0)
        filename = "projected_potential_%.1f_%d.npz" % (pixel_size, thickness)
        numpy.savez(filename, potential=potential, num_atoms=num_atoms)

    # Read the atom data
    atom_data = load_water_atomic_model()

    # Simulate the projected potential
    for pixel_size in numpy.arange(0.1, 2.1, 0.1):
        for thickness in numpy.arange(5, 25, 5):
            print(
                "Compute projected potential for px = %f, dz = %f"
                % (pixel_size, thickness)
            )
            if not os.path.exists(
                "projected_potential_%.1f_%d.npz" % (pixel_size, thickness)
            ):
                compute(atom_data, pixel_size, thickness)


def compute_observed_mean(size, pixel_size):
    """
    Compute the observed mean

    """

    # Create the system configuration
    system_conf = multem.SystemConfiguration()
    system_conf.precision = "float"
    system_conf.device = "device"

    # Create the input multislice configuration
    input_multislice = create_input_multislice()

    # Compute the number of pixels
    nx = int(ceil(size / pixel_size))
    ny = int(ceil(size / pixel_size))
    size = nx * pixel_size

    # Create the specimen atoms
    input_multislice.nx = nx
    input_multislice.ny = ny
    input_multislice.spec_lx = nx * pixel_size
    input_multislice.spec_ly = ny * pixel_size
    input_multislice.spec_lz = nx * pixel_size
    input_multislice.spec_dz = 1

    # For N random placements compute the mean intensity
    means = []
    for j in range(10):

        # Compute the position
        x0 = numpy.random.uniform(0, 1) + nx // 2
        y0 = numpy.random.uniform(0, 1) + ny // 2
        x0 = pixel_size * x0
        y0 = pixel_size * y0

        # Set the atom list
        input_multislice.spec_atoms = multem.AtomList(
            [
                (1, x0, y0, size / 2.0, 0, 1, 0, 0),
                (1, x0, y0, size / 2.0, 0, 1, 0, 0),
                (8, x0, y0, size / 2.0, 0, 1, 0, 0),
            ]
        )

        thickness = []
        potential = []

        def callback(z0, z1, V):
            V = numpy.array(V)
            thickness.append(z1 - z0)
            potential.append(V)

        # Run the simulation
        multem.compute_projected_potential(system_conf, input_multislice, callback)

        # Compute the mean potential
        V = numpy.sum(potential, axis=0)
        means.append(numpy.mean(V))

    # Return the size and mean potential
    return size, numpy.mean(means)


def compute_expected_mean(size):
    """
    Compute the expected mean

    """

    def compute(size, Z):
        hbar = scipy.constants.hbar
        qe = scipy.constants.e
        me = scipy.constants.electron_mass
        # P = multem.compute_V_params("Lobato_0_12", Z, 0)
        P = multem.compute_V_params("Peng_0_12", Z, 0)
        mean = 0
        for i in range(len(P)):
            Ai = P[i][0]
            Bi = P[i][1]
            if Bi > 0:
                k = me * qe / (2 * pi * hbar**2)  # C m^-2 s
                k = 1.0 / (k * 1e-10**2)  # C A^-2 s
                Bj = pi**2 / Bi
                Aj = Ai * k * (Bj / pi) ** (3.0 / 2.0)
                # Bj = (2*pi/Bi)**2
                # Aj = 2*Ai*k*(Bj)**(3.0/2.0)/pi**2
                mean += (1 / k) * Aj
        N = 1
        mean = mean * N / (size**2)
        return mean

    # Return the mean
    return compute(size, 1) * 2 + compute(size, 8)


def compute_mean_correction(ax=None):
    """
    Compute the variance correction table

    """

    size = 40

    pixel_size = numpy.arange(0.1, 2.1, 0.1)
    mean = []
    for ps in pixel_size:
        s, m = compute_observed_mean(size, ps)
        expected = compute_expected_mean(s)
        m = m / expected
        mean.append(m)

    area = [0] + list(pixel_size**2)
    mean = [1] + list(mean)

    # Write out the table to file
    print("Mean0 = %f" % (expected * s**2))
    with open("mean_table.csv", "w") as outfile:
        for x, y in zip(area, mean):
            outfile.write("%.2f, %.7f\n" % (x, y))

    # Plot the mean correction
    ax.plot(area, mean)
    ax.set_xlabel(r"Pixel area ($Å^2$)" + "\n(b)")
    ax.set_title("Mean correction factor")


def compute_variance_correction(ax=None):
    """
    Compute the variance correction table

    """

    X = []
    Y = []

    # Loop through the pixel sizes
    thickness = 20
    for pixel_size in numpy.arange(0.1, 2.1, 0.1):

        # Read the projected potential
        handle = numpy.load("projected_potential_%.1f_%d.npz" % (pixel_size, thickness))
        potential = handle["potential"]
        num_atoms = handle["num_atoms"]

        # Select the pixels with potential in
        xsize, ysize = potential.shape
        x0 = 0  # xsize // 8
        x1 = xsize  # 7 * x0
        y0 = 0  # ysize // 8
        y1 = ysize  # 7 * y0
        potential = potential[x0:x1, y0:y1]

        # Compute the density and variance
        num_molecules = num_atoms / 3.0
        ny, nx = potential.shape
        area = (nx * pixel_size) * (ny * pixel_size)
        var = numpy.var(potential)
        density = num_molecules / (area)

        # Append the pixel area and variance / density
        X.append(pixel_size**2)
        Y.append(var / density)

    X = numpy.array(X)
    Y = numpy.array(Y)

    # Extrapolate to zero and nornalize
    Y0 = Y[0] - X[0] * (Y[1] - Y[0]) / (X[1] - X[0])
    Y = Y / Y0

    X = [0] + list(X)
    Y = [1.0] + list(Y)

    # Write out the table to file
    print("Var0 = %f" % Y0)
    with open("variance_table.csv", "w") as outfile:
        for x, y in zip(X, Y):
            outfile.write("%.2f, %.7f\n" % (x, y))

    # Plot
    ax.plot(X, Y)
    ax.set_xlabel(r"Pixel area ($Å^2$)" + "\n(c)")
    ax.set_title("Variance correction factor")


def compute_power(ax=None):
    """
    Compute the power spectrum

    """

    # Compute the fit to the power spectrum
    pixel_size = 0.1
    for thickness in [20]:  # , 19, 18, 15, 10, 5]:

        # Read the projected potential
        handle = numpy.load("projected_potential_%.1f_%d.npz" % (pixel_size, thickness))
        potential = handle["potential"]
        num_atoms = handle["num_atoms"]

        # Select the pixels with potential in
        xsize, ysize = potential.shape
        x0 = xsize // 8
        x1 = 7 * xsize // 8
        y0 = ysize // 8
        y1 = 7 * ysize // 8
        potential = potential[x0:x1, y0:y1]

        # Compute the density and variance
        num_molecules = num_atoms / 3.0
        ny, nx = potential.shape
        area = (nx * pixel_size) * (ny * pixel_size)
        var = numpy.var(potential)
        mean = numpy.mean(potential)
        # density = num_molecules / (area)
        potential -= mean
        print("Mean: %.3f; Variance: %.3f" % (mean, var))

        # Compute the FFT of the data and the power spectrum
        fft_data = numpy.fft.fft2(potential)
        power = numpy.abs(fft_data) ** 2
        Y, X = numpy.mgrid[0 : power.shape[0], 0 : power.shape[1]]
        q = (1 / pixel_size) * numpy.sqrt(
            ((X - power.shape[1] / 2) / power.shape[1]) ** 2
            + ((Y - power.shape[0] / 2) / power.shape[0]) ** 2
        )
        q = numpy.fft.fftshift(q)

        def func(q, A0, A1, A2, A3):
            M = 1.0 / 2.88
            P = A0 * numpy.exp(-0.5 * q**2 / A1**2) + A2 * numpy.exp(
                -0.5 * (q - M) ** 2 / A3**2
            )
            # I = (2*pi*(A0*A1**2 + A2*M*sqrt(2*pi*A3**2)))
            # P /= I
            model = P
            model[0, 0] = normalized_power[0, 0]
            return model

        # def func2(q, *p):
        #     q = q.reshape(power.shape)
        #     return func(q, *p).flatten()

        # def residuals(p, q, normalized_power):
        #     A0, A1, A2, A3 = p
        #     A0 = numpy.abs(A0)
        #     A2 = numpy.abs(A2)
        #     # A0 = 0.1608
        #     # A2, A3 = 0.822372155, 0.08153797

        #     # A1, A3 = 0.68518427, 0.08693241
        #     M = 1.0 / 2.88
        #     P = A0 * numpy.exp(-0.5 * q ** 2 / A1 ** 2) + A2 * numpy.exp(
        #         -0.5 * (q - M) ** 2 / A3 ** 2
        #     )
        #     # I = (2*pi*(A0*A1**2 + A2*M*sqrt(2*pi*A3**2)))
        #     # P /= I
        #     model = P
        #     model[0, 0] = normalized_power[0, 0]
        #     W = 1.0 / (q * power.shape[0] + 1)
        #     A = numpy.sum(W * (model - normalized_power) ** 2) / numpy.sum(W)
        #     B = (numpy.sum(model) - numpy.sum(normalized_power)) ** 2 / model.size
        #     print(p, A, B)
        #     return A + B

        # Compute the variance correction factor
        Cv = numpy.exp(-3.2056 * pixel_size**2)
        C = Cv * num_molecules / (pixel_size**4)

        # Compute the total integral of the power
        I = numpy.sum(power) * (1 / pixel_size) ** 2 / power.size
        normalized_power = power / I

        # Fit a model to the normalized power
        params = [0.19465002, 0.7312113, 0.78343527, 0.08078005]
        # results = scipy.optimize.minimize(residuals, x0=params, args=(q, normalized_power))#q.flatten(), normalized_power.flatten(), p0=params, sigma=(q*power.shape[0]+1).flatten())
        ##params, _ = scipy.optimize.curve_fit(func2, q.flatten(), normalized_power.flatten(), p0=params, sigma=(q*power.shape[0]+1).flatten())
        # params = results.x
        params[0] = params[0] * I / C
        params[2] = params[2] * I / C
        print("Parameters:", params)

        # Compute the model
        model = func(q, *params)
        model = model * C

        # Compute the radial spectrum
        rp = radial_average(numpy.fft.fftshift(power))
        rm = radial_average(numpy.fft.fftshift(model))
        d = numpy.arange(rp.size) / (pixel_size * power.shape[0])

        # Plot the power spectrum and best fit
        ax.plot(d[1:], rp[1:] / C, label="%d" % thickness)
        ax.plot(d[1:], rm[1:] / C, color="black", alpha=0.5, label="Model")

    # Set some plot properties
    ax.set_xlabel("Spatial frequency ($Å$)\n(a)")
    ax.set_title("Power spectrum")
    ax.set_xlim(0, 1)
    ax.set_yticklabels("")


def calibrate():
    """
    Calibrate ice model

    """
    # Get the water atomic model file
    get_water_atomic_model()

    # Compute the projected potential
    compute_projected_potential()

    # Setup the figure
    width = 0.0393701 * 190
    height = width / 3.0
    fig, ax = pylab.subplots(ncols=3, figsize=(width, height), constrained_layout=True)
    compute_power(ax[0])
    compute_mean_correction(ax[1])
    compute_variance_correction(ax[2])
    fig.savefig("model.png", dpi=300, bbox_inches="tight")


def compute_exit_wave(atom_data, pixel_size):
    """
    Compute the exit wave

    """

    # Get the dimensions
    x_min = atom_data.data["x"].min()
    x_max = atom_data.data["x"].max()
    y_min = atom_data.data["y"].min()
    y_max = atom_data.data["y"].max()
    z_min = atom_data.data["z"].min()
    z_max = atom_data.data["z"].max()
    x_size = x_max - x_min
    y_size = y_max - y_min
    select = (
        (atom_data.data["x"] > x_min + x_size / 6)
        & (atom_data.data["x"] < x_max - x_size / 6)
        & (atom_data.data["y"] > y_min + y_size / 6)
        & (atom_data.data["y"] < y_max - y_size / 6)
    )
    atom_data = parakeet.sample.AtomData(data=atom_data.data[select])
    x_min = atom_data.data["x"].min()
    x_max = atom_data.data["x"].max()
    y_min = atom_data.data["y"].min()
    y_max = atom_data.data["y"].max()
    z_min = atom_data.data["z"].min()
    z_max = atom_data.data["z"].max()
    x_size = x_max - x_min
    y_size = y_max - y_min
    z_size = z_max - z_min

    # Translate to centre
    x_box_size = x_size
    y_box_size = y_size
    z_box_size = z_size

    # Create the system configuration
    system_conf = multem.SystemConfiguration()
    system_conf.precision = "float"
    system_conf.device = "device"

    # Create the input multislice configuration
    input_multislice = create_input_multislice()

    # Compute the number of pixels
    nx = next_power_2(int(x_box_size / pixel_size))
    ny = next_power_2(int(y_box_size / pixel_size))
    assert nx <= 4096
    assert ny <= 4096
    x_box_size = nx * pixel_size
    y_box_size = ny * pixel_size
    x_trans = (x_box_size - x_size) / 2.0 - x_min
    y_trans = (y_box_size - y_size) / 2.0 - y_min
    z_trans = (z_box_size - z_size) / 2.0 - z_min
    atom_data.translate((x_trans, y_trans, z_trans))

    # Create the specimen size
    input_multislice.nx = nx
    input_multislice.ny = ny
    input_multislice.spec_lx = x_box_size
    input_multislice.spec_ly = y_box_size
    input_multislice.spec_lz = z_box_size
    input_multislice.spec_dz = 5

    # Set the specimen atoms
    input_multislice.spec_atoms = atom_data.to_multem()

    # Run the simulation
    output_multislice = multem.simulate(system_conf, input_multislice)

    # Get the image
    physical_image = numpy.array(output_multislice.data[0].psi_coh).T

    # Create the masker
    masker = multem.Masker(input_multislice.nx, input_multislice.ny, pixel_size)

    # Create the size of the cuboid
    masker.set_cuboid(
        (
            x_box_size / 2 - x_size / 2,
            y_box_size / 2 - y_size / 2,
            z_box_size / 2 - z_size / 2,
        ),
        (x_size, y_size, z_size),
    )

    # Run the simulation
    input_multislice.spec_atoms = multem.AtomList()
    output_multislice = multem.simulate(system_conf, input_multislice, masker)

    # Get the image
    random_image = numpy.array(output_multislice.data[0].psi_coh).T

    # Return the images
    x0 = numpy.array((x_box_size / 2 - x_size / 2, y_box_size / 2 - y_size / 2))
    x1 = numpy.array((x_box_size / 2 + x_size / 2, y_box_size / 2 + y_size / 2))
    return physical_image, random_image, x0, x1


def validate():
    """
    Validate the ice model

    """

    # Load the water model
    atom_data = load_water_atomic_model()

    pixel_size = [1.0, 0.1]

    for ps in pixel_size:

        # Get the simulated exit wave
        physical_data, random_data, xmin, xmax = compute_exit_wave(atom_data, ps)

        x0 = numpy.floor(xmin / ps).astype("int32")
        x1 = numpy.floor(xmax / ps).astype("int32")
        xr = x1 - x0
        x0 = x0 + xr // 4
        x1 = x1 - xr // 4

        random_middle = random_data[x0[0] : x1[0], x0[1] : x1[1]]
        physical_middle = physical_data[x0[0] : x1[0], x0[1] : x1[1]]
        physical_middle_mean_real = numpy.mean(physical_middle.flatten().real)
        physical_middle_mean_imag = numpy.mean(physical_middle.flatten().imag)
        random_middle_mean_real = numpy.mean(random_middle.flatten().real)
        random_middle_mean_imag = numpy.mean(random_middle.flatten().imag)

        # pylab.imshow(numpy.abs(random_middle))
        # pylab.show()
        # pylab.imshow(numpy.abs(physical_middle))
        # pylab.show()
        # continue

        physical_middle_std_real = numpy.std(physical_middle.flatten().real)
        physical_middle_std_imag = numpy.std(physical_middle.flatten().imag)
        random_middle_std_real = numpy.std(random_middle.flatten().real)
        random_middle_std_imag = numpy.std(random_middle.flatten().imag)

        # print("Hola")
        width = 0.0393701 * 190
        height = width * 0.75
        fig, ax = pylab.subplots(
            figsize=(width, height),
            nrows=2,
            ncols=2,
            sharex=True,
            sharey=True,
            constrained_layout=True,
        )
        ax[0][0].hist(physical_middle.flatten().real, bins=20, density=True)
        ax[0][1].hist(physical_middle.flatten().imag, bins=20, density=True)
        ax[1][0].hist(random_middle.flatten().real, bins=20, density=True)
        ax[1][1].hist(random_middle.flatten().imag, bins=20, density=True)
        ax[0][0].set_title("Real component", fontweight="bold")
        ax[0][1].set_title("Imaginary component", fontweight="bold")
        ax[0][0].set_ylabel("Physical model", fontweight="bold")
        ax[1][0].set_ylabel("Random model", fontweight="bold")
        ax[0][0].set_xlabel("(a)")
        ax[0][1].set_xlabel("(b)")
        ax[1][0].set_xlabel("(c)")
        ax[1][1].set_xlabel("(d)")
        ax[0][0].axvline(physical_middle_mean_real, color="black")
        ax[0][1].axvline(physical_middle_mean_imag, color="black")
        ax[1][0].axvline(random_middle_mean_real, color="black")
        ax[1][1].axvline(random_middle_mean_imag, color="black")
        # ymax = ax[0][0].get_ylim()[1]
        if ps == 1.0:
            xr = 0.8
            xi = 0.4
        else:
            xr = 0.5
            xi = -1.0
        ax[0][0].text(
            xr,
            0.5 * ax[0][0].get_ylim()[1],
            "mean: %.2f\n sdev: %.2f"
            % (physical_middle_mean_real, physical_middle_std_real),
        )
        ax[0][1].text(
            xi,
            0.5 * ax[0][1].get_ylim()[1],
            "mean: %.2f\n sdev: %.2f"
            % (physical_middle_mean_imag, physical_middle_std_imag),
        )
        ax[1][0].text(
            xr,
            0.5 * ax[1][0].get_ylim()[1],
            "mean: %.2f\n sdev: %.2f"
            % (random_middle_mean_real, random_middle_std_real),
        )
        ax[1][1].text(
            xi,
            0.5 * ax[1][1].get_ylim()[1],
            "mean: %.2f\n sdev: %.2f"
            % (random_middle_mean_imag, random_middle_std_imag),
        )
        fig.savefig("histograms_%.1fA.png" % ps, dpi=300, bbox_inches="tight")

        def compute_power(data, pixel_size):
            f = numpy.fft.fft2(data)
            p = numpy.abs(f) ** 2
            p = numpy.fft.fftshift(p)

            r = radial_average(p)[0 : data.shape[0] // 2]
            d = numpy.arange(r.size) / (pixel_size * data.shape[0])
            return d[1:], r[1:]

        random_d, random_power = compute_power(random_middle, ps)
        physical_d, physical_power = compute_power(physical_middle, ps)

        width = 0.0393701 * 190
        height = width * 0.74
        fig, ax = pylab.subplots(figsize=(width, height), constrained_layout=True)
        ax.plot(physical_d, physical_power, label="Physical model")
        ax.plot(random_d, random_power, label="Random model")
        ax.set_xlabel("Spatial frequency (1/Å)")
        ax.set_ylabel("Power spectrum")
        ax.set_xlim(0, 0.5)
        ax.legend()
        fig.savefig("power_%.1fA.png" % ps, dpi=300, bbox_inches="tight")

        x0 = numpy.floor(xmin / ps).astype("int32")
        # x1 = numpy.floor(xmax / ps).astype("int32")
        x1 = 2 * x0  # + x0 // 2
        x0[:] = 0  # x0 // 2
        random_edge = random_data[x0[0] : x1[0], x0[1] : x1[1]]
        physical_edge = physical_data[x0[0] : x1[0], x0[1] : x1[1]]
        width = 0.0393701 * 190
        height = width
        fig, ax = pylab.subplots(
            figsize=(width, height), ncols=2, constrained_layout=True
        )
        vmin = min(
            numpy.min(numpy.abs(random_edge)), numpy.min(numpy.abs(physical_edge))
        )
        vmax = max(
            numpy.max(numpy.abs(random_edge)), numpy.max(numpy.abs(physical_edge))
        )
        ax[0].imshow(numpy.abs(physical_edge), vmin=vmin, vmax=vmax)
        ax[1].imshow(numpy.abs(random_edge), vmin=vmin, vmax=vmax)
        ax[0].set_title("Physical model", fontweight="bold")
        ax[1].set_title("Random model", fontweight="bold")
        ax[0].set_xticks([])
        ax[1].set_xticks([])
        ax[0].set_yticks([])
        ax[1].set_yticks([])
        ax[0].set_xlabel("(a)")
        ax[1].set_xlabel("(b)")
        fig.savefig("edge_%.1fA.png" % ps, dpi=300, bbox_inches="tight")
        # pylab.show()


if __name__ == "__main__":

    # Create the argument parser
    parser = argparse.ArgumentParser(description="Do the ice model configuration")

    # Add arguments
    parser.add_argument(
        "-c",
        "--calibrate",
        action="store_true",
        default=False,
        dest="calibrate",
        help="Do the calibration",
    )

    parser.add_argument(
        "-v",
        "--validate",
        action="store_true",
        default=False,
        dest="validate",
        help="Do the validation",
    )

    # Parse the arguments
    args = parser.parse_args()

    # Print help
    if not args.calibrate and not args.validate:
        parser.print_help()

    # Do the calibration
    if args.calibrate:
        calibrate()

    # Do the validation
    if args.validate:
        validate()



#
# parakeet.detector.py
#
# Copyright (C) 2019 Diamond Light Source and Rosalind Franklin Institute
#
# Author: James Parkhurst
#
# This code is distributed under the GPLv3 license, a copy of
# which is included in the root directory of this package.
#


class Detector(object):
    """
    A class to encapsulate a detector

    """

    def __init__(self, nx=None, ny=None, pixel_size=None, origin=None, dqe=True):
        """
        Initialise the detector

        Args:
            nx (int): The size of the detector in X
            ny (int): The size of the detector in Y
            pixel_size (float): The effective size of the pixels in A
            origin (tuple): The detector origin (A, A)
            dqe (bool): Use the detector DQE

        """
        self.nx = nx
        self.ny = ny
        self.pixel_size = pixel_size
        self.origin = origin
        self.dqe = dqe


def new(nx=None, ny=None, pixel_size=None, origin=None, dqe=True):
    """
    Make a new detector

    Args:
        nx (int): The size of the detector in X
        ny (int): The size of the detector in Y
        pixel_size (float): The effective size of the pixels in A
        origin (tuple): The detector origin (A, A)
        dqe (bool): Use the detector DQE

    Returns:
        obj: The detector object

    """
    return Detector(nx=nx, ny=ny, pixel_size=pixel_size, origin=origin, dqe=dqe)
